# This file is part of sbi, a toolkit for simulation-based inference. sbi is licensed
# under the Affero General Public License v3, see <https://www.gnu.org/licenses/>.


from abc import ABC
from copy import deepcopy
from typing import Any, Callable, Dict, Optional, Union

import torch
from pyknos.nflows import flows
from torch import Tensor, nn, optim
from torch.distributions import Distribution
from torch.nn.utils.clip_grad import clip_grad_norm_
from torch.utils import data
from torch.utils.tensorboard.writer import SummaryWriter

from sbi import utils as utils
from sbi.inference import NeuralInference
from sbi.inference.posteriors import MCMCPosterior, RejectionPosterior, VIPosterior
from sbi.inference.potentials import likelihood_estimator_based_potential
from sbi.utils import (
    check_estimator_arg,
    check_prior,
    mask_sims_from_prior,
    validate_theta_and_x,
    x_shape_from_simulation,
)


class LikelihoodEstimator(NeuralInference, ABC):
    def __init__(
        self,
        prior: Optional[Distribution] = None,
        density_estimator: Union[str, Callable] = "maf",
        device: str = "cpu",
        logging_level: Union[int, str] = "WARNING",
        summary_writer: Optional[SummaryWriter] = None,
        show_progress_bars: bool = True,
    ):
        r"""Base class for Sequential Neural Likelihood Estimation methods.

        Args:
            prior: A probability distribution that expresses prior knowledge about the
                parameters, e.g. which ranges are meaningful for them. Any
                object with `.log_prob()`and `.sample()` (for example, a PyTorch
                distribution) can be used.
            density_estimator: If it is a string, use a pre-configured network of the
                provided type (one of nsf, maf, mdn, made). Alternatively, a function
                that builds a custom neural network can be provided. The function will
                be called with the first batch of simulations (theta, x), which can
                thus be used for shape inference and potentially for z-scoring. It
                needs to return a PyTorch `nn.Module` implementing the density
                estimator. The density estimator needs to provide the methods
                `.log_prob` and `.sample()`.

        See docstring of `NeuralInference` class for all other arguments.
        """

        super().__init__(
            prior=prior,
            device=device,
            logging_level=logging_level,
            summary_writer=summary_writer,
            show_progress_bars=show_progress_bars,
        )

        # As detailed in the docstring, `density_estimator` is either a string or
        # a callable. The function creating the neural network is attached to
        # `_build_neural_net`. It will be called in the first round and receive
        # thetas and xs as inputs, so that they can be used for shape inference and
        # potentially for z-scoring.
        check_estimator_arg(density_estimator)
        if isinstance(density_estimator, str):
            self._build_neural_net = utils.likelihood_nn(model=density_estimator)
        else:
            self._build_neural_net = density_estimator

        # SNLE-specific summary_writer fields.
        self._summary.update({"mcmc_times": []})  # type: ignore

    def append_simulations(
        self,
        theta: Tensor,
        x: Tensor,
        from_round: int = 0,
    ) -> "LikelihoodEstimator":
        r"""Store parameters and simulation outputs to use them for later training.

        Data are stored as entries in lists for each type of variable (parameter/data).

        Stores $\theta$, $x$, prior_masks (indicating if simulations are coming from the
        prior or not) and an index indicating which round the batch of simulations came
        from.

        Args:
            theta: Parameter sets.
            x: Simulation outputs.
            from_round: Which round the data stemmed from. Round 0 means from the prior.
                With default settings, this is not used at all for `SNLE`. Only when
                the user later on requests `.train(discard_prior_samples=True)`, we
                use these indices to find which training data stemmed from the prior.

        Returns:
            NeuralInference object (returned so that this function is chainable).
        """

        theta, x = validate_theta_and_x(theta, x, training_device=self._device)

        self._theta_roundwise.append(theta)
        self._x_roundwise.append(x)
        self._prior_masks.append(mask_sims_from_prior(int(from_round), theta.size(0)))
        self._data_round_index.append(int(from_round))

        return self

    def train(
        self,
        training_batch_size: int = 50,
        learning_rate: float = 5e-4,
        validation_fraction: float = 0.1,
        stop_after_epochs: int = 20,
        max_num_epochs: int = 2**31 - 1,
        clip_max_norm: Optional[float] = 5.0,
        exclude_invalid_x: bool = True,
        resume_training: bool = False,
        discard_prior_samples: bool = False,
        retrain_from_scratch: bool = False,
        show_train_summary: bool = False,
        dataloader_kwargs: Optional[Dict] = None,
    ) -> flows.Flow:
        r"""Train the density estimator to learn the distribution $p(x|\theta)$.

        Args:
            exclude_invalid_x: Whether to exclude simulation outputs `x=NaN` or `x=±∞`
                during training. Expect errors, silent or explicit, when `False`.
            resume_training: Can be used in case training time is limited, e.g. on a
                cluster. If `True`, the split between train and validation set, the
                optimizer, the number of epochs, and the best validation log-prob will
                be restored from the last time `.train()` was called.
            discard_prior_samples: Whether to discard samples simulated in round 1, i.e.
                from the prior. Training may be sped up by ignoring such less targeted
                samples.
            retrain_from_scratch: Whether to retrain the conditional density
                estimator for the posterior from scratch each round.
            show_train_summary: Whether to print the number of epochs and validation
                loss after the training.
            dataloader_kwargs: Additional or updated kwargs to be passed to the training
                and validation dataloaders (like, e.g., a collate_fn)

        Returns:
            Density estimator that has learned the distribution $p(x|\theta)$.
        """

        # Starting index for the training set (1 = discard round-0 samples).
        start_idx = int(discard_prior_samples and self._round > 0)
        # Load data from most recent round.
        self._round = max(self._data_round_index)
        theta, x, _ = self.get_simulations(
            start_idx, exclude_invalid_x, warn_on_invalid=True
        )

        # Dataset is shared for training and validation loaders.
        dataset = data.TensorDataset(theta, x)

        train_loader, val_loader = self.get_dataloaders(
            dataset,
            training_batch_size,
            validation_fraction,
            resume_training,
            dataloader_kwargs=dataloader_kwargs,
        )

        # First round or if retraining from scratch:
        # Call the `self._build_neural_net` with the rounds' thetas and xs as
        # arguments, which will build the neural network
        # This is passed into NeuralPosterior, to create a neural posterior which
        # can `sample()` and `log_prob()`. The network is accessible via `.net`.
        if self._neural_net is None or retrain_from_scratch:
            self._neural_net = self._build_neural_net(
                theta[self.train_indices], x[self.train_indices]
            )
            self._x_shape = x_shape_from_simulation(x)
            assert (
                len(self._x_shape) < 3
            ), "SNLE cannot handle multi-dimensional simulator output."

        self._neural_net.to(self._device)
        if not resume_training:
            self.optimizer = optim.Adam(
                list(self._neural_net.parameters()),
                lr=learning_rate,
            )
            self.epoch, self._val_log_prob = 0, float("-Inf")

        while self.epoch <= max_num_epochs and not self._converged(
            self.epoch, stop_after_epochs
        ):

            # Train for a single epoch.
            self._neural_net.train()
            train_log_probs_sum = 0
            for batch in train_loader:
                self.optimizer.zero_grad()
                theta_batch, x_batch = (
                    batch[0].to(self._device),
                    batch[1].to(self._device),
                )
                # Evaluate on x with theta as context.
                train_losses = self._loss(theta=theta_batch, x=x_batch)
                train_loss = torch.mean(train_losses)
                train_log_probs_sum -= train_losses.sum().item()

                train_loss.backward()
                if clip_max_norm is not None:
                    clip_grad_norm_(
                        self._neural_net.parameters(),
                        max_norm=clip_max_norm,
                    )
                self.optimizer.step()

            self.epoch += 1

            train_log_prob_average = train_log_probs_sum / (
                len(train_loader) * train_loader.batch_size  # type: ignore
            )
            self._summary["train_log_probs"].append(train_log_prob_average)

            # Calculate validation performance.
            self._neural_net.eval()
            val_log_prob_sum = 0
            with torch.no_grad():
                for batch in val_loader:
                    theta_batch, x_batch = (
                        batch[0].to(self._device),
                        batch[1].to(self._device),
                    )
                    # Evaluate on x with theta as context.
                    val_losses = self._loss(theta=theta_batch, x=x_batch)
                    val_log_prob_sum -= val_losses.sum().item()

            # Take mean over all validation samples.
            self._val_log_prob = val_log_prob_sum / (
                len(val_loader) * val_loader.batch_size  # type: ignore
            )
            # Log validation log prob for every epoch.
            self._summary["validation_log_probs"].append(self._val_log_prob)

            self._maybe_show_progress(self._show_progress_bars, self.epoch)

        self._report_convergence_at_end(self.epoch, stop_after_epochs, max_num_epochs)

        # Update summary.
        self._summary["epochs"].append(self.epoch)
        self._summary["best_validation_log_probs"].append(self._best_val_log_prob)

        # Update TensorBoard and summary dict.
        self._summarize(
            round_=self._round,
            x_o=None,
            theta_bank=theta,
            x_bank=x,
        )

        # Update description for progress bar.
        if show_train_summary:
            print(self._describe_round(self._round, self._summary))

        # Avoid keeping the gradients in the resulting network, which can
        # cause memory leakage when benchmarking.
        self._neural_net.zero_grad(set_to_none=True)

        return deepcopy(self._neural_net)

    def build_posterior(
        self,
        density_estimator: Optional[nn.Module] = None,
        prior: Optional[Distribution] = None,
        sample_with: str = "mcmc",
        mcmc_method: str = "slice_np",
        vi_method: str = "rKL",
        mcmc_parameters: Dict[str, Any] = {},
        vi_parameters: Dict[str, Any] = {},
        rejection_sampling_parameters: Dict[str, Any] = {},
    ) -> Union[MCMCPosterior, RejectionPosterior, VIPosterior]:
        r"""Build posterior from the neural density estimator.

        SNLE trains a neural network to approximate the likelihood $p(x|\theta)$. The
        posterior wraps the trained network such that one can directly evaluate the
        unnormalized posterior log probability $p(\theta|x) \propto p(x|\theta) \cdot
        p(\theta)$ and draw samples from the posterior with MCMC or rejection sampling.

        Args:
            density_estimator: The density estimator that the posterior is based on.
                If `None`, use the latest neural density estimator that was trained.
            prior: Prior distribution.
            sample_with: Method to use for sampling from the posterior. Must be one of
                [`mcmc` | `rejection` | `vi`].
            mcmc_method: Method used for MCMC sampling, one of `slice_np`, `slice`,
                `hmc`, `nuts`. Currently defaults to `slice_np` for a custom numpy
                implementation of slice sampling; select `hmc`, `nuts` or `slice` for
                Pyro-based sampling.
            vi_method: Method used for VI, one of [`rKL`, `fKL`, `IW`, `alpha`]. Note
                some of the methods admit a `mode seeking` property (e.g. rKL) whereas
                some admit a `mass covering` one (e.g fKL).
            mcmc_parameters: Additional kwargs passed to `MCMCPosterior`.
            vi_parameters: Additional kwargs passed to `VIPosterior`.
            rejection_sampling_parameters: Additional kwargs passed to
                `RejectionPosterior`.

        Returns:
            Posterior $p(\theta|x)$  with `.sample()` and `.log_prob()` methods
            (the returned log-probability is unnormalized).
        """
        if prior is None:
            assert (
                self._prior is not None
            ), """You did not pass a prior. You have to pass the prior either at
            initialization `inference = SNLE(prior)` or to `.build_posterior
            (prior=prior)`."""
            prior = self._prior
        else:
            check_prior(prior)

        if density_estimator is None:
            likelihood_estimator = self._neural_net
            # If internal net is used device is defined.
            device = self._device
        else:
            likelihood_estimator = density_estimator
            # Otherwise, infer it from the device of the net parameters.
            device = next(density_estimator.parameters()).device.type

        potential_fn, theta_transform = likelihood_estimator_based_potential(
            likelihood_estimator=likelihood_estimator, prior=prior, x_o=None
        )

        if sample_with == "mcmc":
            self._posterior = MCMCPosterior(
                potential_fn=potential_fn,
                theta_transform=theta_transform,
                proposal=prior,
                method=mcmc_method,
                device=device,
                x_shape=self._x_shape,
                **mcmc_parameters,
            )
        elif sample_with == "rejection":
            self._posterior = RejectionPosterior(
                potential_fn=potential_fn,
                proposal=prior,
                device=device,
                x_shape=self._x_shape,
                **rejection_sampling_parameters,
            )
        elif sample_with == "vi":
            self._posterior = VIPosterior(
                potential_fn=potential_fn,
                theta_transform=theta_transform,
                prior=prior,  # type: ignore
                vi_method=vi_method,
                device=device,
                x_shape=self._x_shape,
                **vi_parameters,
            )
        else:
            raise NotImplementedError

        # Store models at end of each round.
        self._model_bank.append(deepcopy(self._posterior))

        return deepcopy(self._posterior)

    def _loss(self, theta: Tensor, x: Tensor) -> Tensor:
        r"""Return loss for SNLE, which is the likelihood of $-\log q(x_i | \theta_i)$.

        Returns:
            Negative log prob.
        """
        return -self._neural_net.log_prob(x, context=theta)



# This file is part of sbi, a toolkit for simulation-based inference. sbi is licensed
# under the Affero General Public License v3, see <https://www.gnu.org/licenses/>.

from __future__ import annotations

from contextlib import nullcontext
from typing import Optional, Tuple

import numpy as np
import pytest
import torch
import torch.distributions.transforms as torch_tf
import torch.nn as nn
from torch import eye, ones, zeros
from torch.distributions import Distribution, MultivariateNormal

from sbi import utils as utils
from sbi.inference import (
    SNLE,
    SNPE_A,
    SNPE_C,
    SNRE_A,
    SNRE_B,
    DirectPosterior,
    MCMCPosterior,
    NeuralInference,
    RejectionPosterior,
    VIPosterior,
    likelihood_estimator_based_potential,
    posterior_estimator_based_potential,
    ratio_estimator_based_potential,
    simulate_for_sbi,
)
from sbi.inference.potentials.base_potential import BasePotential
from sbi.simulators import diagonal_linear_gaussian, linear_gaussian
from sbi.utils.get_nn_models import classifier_nn, likelihood_nn, posterior_nn
from sbi.utils.torchutils import BoxUniform, process_device
from sbi.utils.user_input_checks import (
    check_embedding_net_device,
    prepare_for_sbi,
    validate_theta_and_x,
)


@pytest.mark.slow
@pytest.mark.gpu
@pytest.mark.parametrize(
    "method, model, mcmc_method",
    [
        (SNPE_C, "mdn", "rejection"),
        (SNPE_C, "maf", "slice"),
        (SNPE_C, "maf", "direct"),
        (SNLE, "maf", "slice"),
        (SNLE, "nsf", "slice_np"),
        (SNLE, "nsf", "rejection"),
        (SNRE_A, "mlp", "slice_np_vectorized"),
        (SNRE_B, "resnet", "nuts"),
        (SNRE_B, "resnet", "rejection"),
    ],
)
@pytest.mark.parametrize("data_device", ("cpu", "cuda:0"))
@pytest.mark.parametrize(
    "training_device, prior_device",
    [
        pytest.param("cpu", "cuda", marks=pytest.mark.xfail),
        pytest.param("cuda:0", "cpu", marks=pytest.mark.xfail),
        ("cuda:0", "cuda:0"),
        ("cuda:0", "cuda:0"),
        ("cpu", "cpu"),
    ],
)
def test_training_and_mcmc_on_device(
    method,
    model,
    data_device,
    mcmc_method,
    training_device,
    prior_device,
    prior_type="gaussian",
):
    """Test training on devices.

    This test does not check training speeds.

    """

    num_dim = 2
    num_samples = 10
    num_simulations = 100
    max_num_epochs = 5

    x_o = zeros(1, num_dim).to(data_device)
    likelihood_shift = -1.0 * ones(num_dim).to(prior_device)
    likelihood_cov = 0.3 * eye(num_dim).to(prior_device)

    if prior_type == "gaussian":
        prior_mean = zeros(num_dim).to(prior_device)
        prior_cov = eye(num_dim).to(prior_device)
        prior = MultivariateNormal(loc=prior_mean, covariance_matrix=prior_cov)
    else:
        prior = BoxUniform(
            low=-2 * torch.ones(num_dim),
            high=2 * torch.ones(num_dim),
            device=prior_device,
        )

    def simulator(theta):
        return linear_gaussian(theta, likelihood_shift, likelihood_cov)

    training_device = process_device(training_device)

    if method in [SNPE_A, SNPE_C]:
        kwargs = dict(
            density_estimator=utils.posterior_nn(model=model, num_transforms=2)
        )
    elif method == SNLE:
        kwargs = dict(
            density_estimator=utils.likelihood_nn(model=model, num_transforms=2)
        )
    elif method in (SNRE_A, SNRE_B):
        kwargs = dict(classifier=utils.classifier_nn(model=model))
    else:
        raise ValueError()

    inferer = method(show_progress_bars=False, device=training_device, **kwargs)

    proposals = [prior]

    # Test for two rounds.
    for _ in range(2):
        theta, x = simulate_for_sbi(simulator, proposals[-1], num_simulations)
        theta, x = theta.to(data_device), x.to(data_device)

        estimator = inferer.append_simulations(theta, x).train(
            training_batch_size=100, max_num_epochs=max_num_epochs
        )
        if method == SNLE:
            potential_fn, theta_transform = likelihood_estimator_based_potential(
                estimator, prior, x_o
            )
        elif method == SNPE_A or method == SNPE_C:
            potential_fn, theta_transform = posterior_estimator_based_potential(
                estimator, prior, x_o
            )
        elif method == SNRE_A or method == SNRE_B:
            potential_fn, theta_transform = ratio_estimator_based_potential(
                estimator, prior, x_o
            )
        else:
            raise ValueError

        if mcmc_method == "rejection":
            posterior = RejectionPosterior(
                proposal=prior,
                potential_fn=potential_fn,
                device=training_device,
            )
        elif mcmc_method == "direct":
            posterior = DirectPosterior(
                posterior_estimator=estimator, prior=prior
            ).set_default_x(x_o)
        else:
            posterior = MCMCPosterior(
                potential_fn=potential_fn,
                theta_transform=theta_transform,
                proposal=prior,
                method=mcmc_method,
                device=training_device,
            )
        proposals.append(posterior)

    # Check for default device for inference object
    weights_device = next(inferer._neural_net.parameters()).device
    assert torch.device(training_device) == weights_device
    samples = proposals[-1].sample(sample_shape=(num_samples,))
    proposals[-1].potential(samples)


@pytest.mark.gpu
@pytest.mark.parametrize(
    "device_input, device_target",
    [
        ("cpu", "cpu"),
        ("cuda", "cuda:0"),
        ("cuda:0", "cuda:0"),
        pytest.param("cuda:42", None, marks=pytest.mark.xfail),
        pytest.param("qwerty", None, marks=pytest.mark.xfail),
    ],
)
def test_process_device(device_input: str, device_target: Optional[str]) -> None:
    device_output = process_device(device_input)
    assert device_output == device_target, (
        f"Failure when processing device '{device_input}': "
        f"result should have been '{device_target}' and is "
        f"instead '{device_output}'"
    )


@pytest.mark.gpu
@pytest.mark.parametrize("device_datum", ["cpu", "cuda"])
@pytest.mark.parametrize("device_embedding_net", ["cpu", "cuda"])
def test_check_embedding_net_device(
    device_datum: str, device_embedding_net: str
) -> None:

    datum = torch.zeros((1, 1)).to(device_datum)
    embedding_net = nn.Linear(in_features=1, out_features=1).to(device_embedding_net)

    if device_datum != device_embedding_net:
        with pytest.warns(UserWarning):
            check_embedding_net_device(datum=datum, embedding_net=embedding_net)
    else:
        check_embedding_net_device(datum=datum, embedding_net=embedding_net)

    output_device_net = [p.device for p in embedding_net.parameters()][0]
    assert datum.device == output_device_net, (
        f"Failure when processing embedding_net: "
        f"device should have been set to should have been '{datum.device}' but is "
        f"still '{output_device_net}'"
    )


@pytest.mark.parametrize(
    "shape_x",
    [
        (3, 1),
    ],
)
@pytest.mark.parametrize(
    "shape_theta", [(3, 2), pytest.param((2, 1), marks=pytest.mark.xfail)]
)
def test_validate_theta_and_x_shapes(
    shape_x: Tuple[int], shape_theta: Tuple[int]
) -> None:
    x = torch.empty(shape_x)
    theta = torch.empty(shape_theta)

    validate_theta_and_x(theta, x, training_device="cpu")


def test_validate_theta_and_x_tensor() -> None:
    x = torch.empty((1, 1))
    theta = torch.ones((1, 1)).tolist()

    with pytest.raises(Exception):
        validate_theta_and_x(theta, x, training_device="cpu")


def test_validate_theta_and_x_type() -> None:
    x = torch.empty((1, 1))
    theta = torch.empty((1, 1), dtype=int)

    with pytest.raises(Exception):
        validate_theta_and_x(theta, x, training_device="cpu")


@pytest.mark.gpu
@pytest.mark.parametrize("training_device", ["cpu", "cuda:0"])
@pytest.mark.parametrize("data_device", ["cpu", "cuda:0"])
def test_validate_theta_and_x_device(training_device: str, data_device: str) -> None:
    theta = torch.empty((1, 1)).to(data_device)
    x = torch.empty((1, 1)).to(data_device)

    if training_device != data_device:
        with pytest.warns(UserWarning):
            theta, x = validate_theta_and_x(theta, x, training_device=training_device)
    else:
        theta, x = validate_theta_and_x(theta, x, training_device=training_device)

    assert str(theta.device) == training_device, (
        f"Data should have its device converted from '{data_device}' "
        f"to training_device '{training_device}'."
    )


@pytest.mark.gpu
@pytest.mark.parametrize("inference_method", [SNPE_A, SNPE_C, SNRE_A, SNRE_B, SNLE])
@pytest.mark.parametrize("data_device", ("cpu", "cuda:0"))
@pytest.mark.parametrize("training_device", ("cpu", "cuda:0"))
def test_train_with_different_data_and_training_device(
    inference_method, data_device: str, training_device: str
) -> None:

    assert torch.cuda.is_available(), "this test requires that cuda is available."

    num_dim = 2
    prior_ = BoxUniform(
        -torch.ones(num_dim), torch.ones(num_dim), device=training_device
    )
    simulator, prior = prepare_for_sbi(diagonal_linear_gaussian, prior_)

    inference = inference_method(
        prior,
        **(
            dict(classifier="resnet")
            if inference_method in [SNRE_A, SNRE_B]
            else dict(
                density_estimator=(
                    "mdn_snpe_a" if inference_method == SNPE_A else "maf"
                )
            )
        ),
        show_progress_bars=False,
        device=training_device,
    )

    theta, x = simulate_for_sbi(simulator, prior, 32)
    theta, x = theta.to(data_device), x.to(data_device)
    x_o = torch.zeros(x.shape[1])
    inference = inference.append_simulations(theta, x)

    posterior_estimator = inference.train(max_num_epochs=2)

    # Check for default device for inference object
    weights_device = next(inference._neural_net.parameters()).device
    assert torch.device(training_device) == weights_device

    _ = DirectPosterior(
        posterior_estimator=posterior_estimator, prior=prior
    ).set_default_x(x_o)


@pytest.mark.gpu
@pytest.mark.parametrize("inference_method", [SNPE_A, SNPE_C, SNRE_A, SNRE_B, SNLE])
@pytest.mark.parametrize("prior_device", ("cpu", "cuda"))
@pytest.mark.parametrize("embedding_net_device", ("cpu", "cuda"))
@pytest.mark.parametrize("data_device", ("cpu", "cuda"))
@pytest.mark.parametrize("training_device", ("cpu", "cuda"))
def test_embedding_nets_integration_training_device(
    inference_method,
    prior_device: str,
    embedding_net_device: str,
    data_device: str,
    training_device: str,
) -> None:

    # add other methods

    D_theta = 2
    D_x = 3
    samples_per_round = 32
    num_rounds = 2

    x_o = torch.ones((1, D_x))

    prior = utils.BoxUniform(
        low=-torch.ones((D_theta,)), high=torch.ones((D_theta,)), device=prior_device
    )

    if inference_method in [SNRE_A, SNRE_B]:
        embedding_net_theta = nn.Linear(in_features=D_theta, out_features=2).to(
            embedding_net_device
        )
        embedding_net_x = nn.Linear(in_features=D_x, out_features=2).to(
            embedding_net_device
        )
        nn_kwargs = dict(
            classifier=classifier_nn(
                model="resnet",
                embedding_net_x=embedding_net_x,
                embedding_net_theta=embedding_net_theta,
                hidden_features=4,
            )
        )
    elif inference_method == SNLE:
        embedding_net = nn.Linear(in_features=D_theta, out_features=2).to(
            embedding_net_device
        )
        nn_kwargs = dict(
            density_estimator=likelihood_nn(
                model="maf",
                embedding_net=embedding_net,
                hidden_features=4,
                num_transforms=2,
            )
        )
    else:
        embedding_net = nn.Linear(in_features=D_x, out_features=2).to(
            embedding_net_device
        )
        nn_kwargs = dict(
            density_estimator=posterior_nn(
                model="mdn_snpe_a" if inference_method == SNPE_A else "maf",
                embedding_net=embedding_net,
                hidden_features=4,
                num_transforms=2,
            )
        )

    with pytest.raises(Exception) if prior_device != training_device else nullcontext():
        inference = inference_method(prior=prior, **nn_kwargs, device=training_device)

    if prior_device != training_device:
        pytest.xfail("We do not correct the case of invalid prior device")

    theta = prior.sample((samples_per_round,)).to(data_device)

    proposal = prior
    for round_idx in range(num_rounds):
        X = (
            MultivariateNormal(torch.zeros((D_x,)), torch.eye(D_x))
            .sample((samples_per_round,))
            .to(data_device)
        )

        with pytest.warns(
            UserWarning
        ) if data_device != training_device else nullcontext():
            density_estimator_append = inference.append_simulations(theta, X)

        with pytest.warns(UserWarning) if (round_idx == 0) and (
            embedding_net_device != training_device
        ) else nullcontext():
            density_estimator_train = density_estimator_append.train(max_num_epochs=2)

        posterior = inference.build_posterior(density_estimator_train)
        proposal = posterior.set_default_x(x_o)
        theta = proposal.sample((samples_per_round,))


@pytest.mark.parametrize("inference_method", [SNPE_A, SNPE_C, SNRE_A, SNRE_B, SNLE])
def test_nograd_after_inference_train(inference_method) -> None:

    num_dim = 2
    prior_ = BoxUniform(-torch.ones(num_dim), torch.ones(num_dim))
    simulator, prior = prepare_for_sbi(diagonal_linear_gaussian, prior_)

    inference = inference_method(
        prior,
        **(
            dict(classifier="resnet")
            if inference_method in [SNRE_A, SNRE_B]
            else dict(
                density_estimator=(
                    "mdn_snpe_a" if inference_method == SNPE_A else "maf"
                )
            )
        ),
        show_progress_bars=False,
    )

    theta, x = simulate_for_sbi(simulator, prior, 32)
    inference = inference.append_simulations(theta, x)

    posterior_estimator = inference.train(max_num_epochs=2)

    def check_no_grad(model):
        for p in model.parameters():
            assert p.grad is None

    check_no_grad(posterior_estimator)
    check_no_grad(inference._neural_net)


@pytest.mark.slow
@pytest.mark.gpu
@pytest.mark.parametrize("num_dim", (1, 2))
@pytest.mark.parametrize("q", ("maf", "nsf", "gaussian_diag", "gaussian", "mcf", "scf"))
@pytest.mark.parametrize("vi_method", ("rKL", "fKL", "IW", "alpha"))
@pytest.mark.parametrize("sampling_method", ("naive", "sir"))
def test_vi_on_gpu(num_dim: int, q: Distribution, vi_method: str, sampling_method: str):
    """Test VI on Gaussian, comparing to ground truth target via c2st.

    Args:
        num_dim: parameter dimension of the gaussian model
        vi_method: different vi methods
        sampling_method: Different sampling methods

    """

    device = "cuda:0"

    if num_dim == 1 and q in ["mcf", "scf"]:
        return

    # Good run where everythink is one the correct device.
    class FakePotential(BasePotential):
        def __call__(self, theta, **kwargs):
            return torch.ones(len(theta), dtype=torch.float32, device=device)

        def allow_iid_x(self) -> bool:
            return True

    potential_fn = FakePotential(
        prior=MultivariateNormal(
            zeros(num_dim, device=device), eye(num_dim, device=device)
        ),
        device=device,
    )
    theta_transform = torch_tf.identity_transform

    posterior = VIPosterior(
        potential_fn=potential_fn, theta_transform=theta_transform, q=q, device=device
    )
    posterior.set_default_x(
        torch.tensor(np.zeros((num_dim,)).astype(np.float32)).to(device)
    )
    posterior.vi_method = vi_method

    posterior.train(min_num_iters=9, max_num_iters=10, warm_up_rounds=10)
    samples = posterior.sample((1,), method=sampling_method)
    logprobs = posterior.log_prob(samples)

    assert str(samples.device) == device, "The devices after training does not match"
    assert str(logprobs.device) == device, "The devices after training does not match"



# This file is part of sbi, a toolkit for simulation-based inference. sbi is licensed
# under the Affero General Public License v3, see <https://www.gnu.org/licenses/>.

from typing import Any, Dict, Optional

import numpy as np
import torch
from sklearn.ensemble import RandomForestClassifier
from sklearn.model_selection import KFold, cross_val_score
from sklearn.neural_network import MLPClassifier
from torch import Tensor


def c2st(
    X: Tensor,
    Y: Tensor,
    seed: int = 1,
    n_folds: int = 5,
    metric: str = "accuracy",
    classifier: str = "rf",
) -> Tensor:
    """
    Return accuracy of classifier trained to distinguish samples from supposedly
    two distributions <X> and <Y>. For details on the method, see [1,2].
    If the returned accuracy is 0.5, <X> and <Y> are considered to be from the
    same generating PDF, i.e. they can not be differentiated.
    If the returned accuracy is around 1., <X> and <Y> are considered to be from
    two different generating PDFs.

    Training of the classifier with N-fold cross-validation [3] using sklearn.
    By default, a `RandomForestClassifier` by from `sklearn.ensemble` is used
    (<classifier> = 'rf'). Alternatively, a multi-layer perceptron is available
    (<classifier> = 'mlp'). For a small study on the pros and cons for this
    choice see [4]. Before both samples are ingested, they are normalized (z scored)
    under the assumption that each dimension in X follows a normal distribution, i.e.
    the mean(X) is subtracted from X and this difference is divided by std(X)
    for every dimension.

    If you need a more flexible interface which is able to take a sklearn
    compatible classifier and more, see the `c2st_` method in this module.

    Args:
        X: Samples from one distribution.
        Y: Samples from another distribution.
        seed: Seed for the sklearn classifier and the KFold cross-validation
        n_folds: Number of folds to use
        metric: sklearn compliant metric to use for the scoring parameter of cross_val_score
        classifier: classification architecture to use, possible values: 'rf' or 'mlp'

    Return:
        torch.tensor containing the mean accuracy score over the test sets
        from cross-validation

    Example:
    ``` py
    > c2st(X,Y)
    [0.51904464] #X and Y likely come from the same PDF or ensemble
    > c2st(P,Q)
    [0.998456] #P and Q likely come from two different PDFs or ensembles
    ```

    References:
        [1]: http://arxiv.org/abs/1610.06545
        [2]: https://www.osti.gov/biblio/826696/
        [3]: https://scikit-learn.org/stable/modules/cross_validation.html
        [4]: https://github.com/psteinb/c2st/
    """

    # the default configuration
    clf_class = RandomForestClassifier
    clf_kwargs = {}

    if "mlp" in classifier.lower():
        ndim = X.shape[-1]
        clf_class = MLPClassifier
        clf_kwargs = {
            "activation": "relu",
            "hidden_layer_sizes": (10 * ndim, 10 * ndim),
            "max_iter": 1000,
            "solver": "adam",
            "early_stopping": True,
            "n_iter_no_change": 50,
        }

    noise_scale = None
    z_score = True
    verbosity = 0

    scores_ = c2st_scores(
        X,
        Y,
        seed=seed,
        n_folds=n_folds,
        metric=metric,
        z_score=z_score,
        noise_scale=noise_scale,
        verbosity=verbosity,
        clf_class=clf_class,
        clf_kwargs=clf_kwargs,
    )

    # TODO: unclear why np.asarray needs to be used here
    scores = np.asarray(np.mean(scores_)).astype(np.float32)
    value = torch.from_numpy(np.atleast_1d(scores))
    return value


def c2st_scores(
    X: Tensor,
    Y: Tensor,
    seed: int = 1,
    n_folds: int = 5,
    metric: str = "accuracy",
    z_score: bool = True,
    noise_scale: Optional[float] = None,
    verbosity: int = 0,
    clf_class: Any = RandomForestClassifier,
    clf_kwargs: Dict[str, Any] = {},
) -> Tensor:
    """
    Return accuracy of classifier trained to distinguish samples from supposedly
    two distributions <X> and <Y>. For details on the method, see [1,2].
    If the returned accuracy is 0.5, <X> and <Y> are considered to be from the
    same generating PDF, i.e. they can not be differentiated.
    If the returned accuracy is around 1., <X> and <Y> are considered to be from
    two different generating PDFs.

    This function performs training of the classifier with N-fold cross-validation [3] using sklearn.
    By default, a `RandomForestClassifier` by from `sklearn.ensemble` is used which
    is recommended based on the study performed in [4].
    This can be changed using <clf_class>. This class is used in the following
    fashion:

    ``` py
    clf = clf_class(random_state=seed, **clf_kwargs)
    #...
    scores = cross_val_score(
        clf, data, target, cv=shuffle, scoring=scoring, verbose=verbosity
    )
    ```
    Further configuration of the classifier can be performed using <clf_kwargs>.
    If you like to provide a custom class for training, it has to satisfy the
    internal requirements of `sklearn.model_selection.cross_val_score`.

    Args:
        X: Samples from one distribution.
        Y: Samples from another distribution.
        seed: Seed for the sklearn classifier and the KFold cross validation
        n_folds: Number of folds to use for cross validation
        metric: sklearn compliant metric to use for the scoring parameter of cross_val_score
        z_score: Z-scoring using X, i.e. mean and std deviation of X is used to normalize Y, i.e. Y=(Y - mean)/std
        noise_scale: If passed, will add Gaussian noise with standard deviation <noise_scale> to samples of X and of Y
        verbosity: control the verbosity of sklearn.model_selection.cross_val_score
        clf_class: a scikit-learn classifier class
        clf_kwargs: key-value arguments dictionary to the class specified by clf_class, e.g. sklearn.ensemble.RandomForestClassifier

    Return:
        np.ndarray containing the calculated <metric> scores over the test set
        folds from cross-validation

    Example:
    ``` py
    > c2st_scores(X,Y)
    [0.51904464,0.5309201,0.4959452,0.5487709,0.50682926]
    #X and Y likely come from the same PDF or ensemble
    > c2st_scores(P,Q)
    [0.998456,0.9982912,0.9980476,0.9980488,0.99805826]
    #P and Q likely come from two different PDFs or ensembles
    ```

    References:
        [1]: http://arxiv.org/abs/1610.06545
        [2]: https://www.osti.gov/biblio/826696/
        [3]: https://scikit-learn.org/stable/modules/cross_validation.html
        [4]: https://github.com/psteinb/c2st/
    """
    if z_score:
        X_mean = torch.mean(X, dim=0)
        X_std = torch.std(X, dim=0)
        X = (X - X_mean) / X_std
        Y = (Y - X_mean) / X_std

    if noise_scale is not None:
        X += noise_scale * torch.randn(X.shape)
        Y += noise_scale * torch.randn(Y.shape)

    X = X.cpu().numpy()
    Y = Y.cpu().numpy()

    clf = clf_class(random_state=seed, **clf_kwargs)

    # prepare data
    data = np.concatenate((X, Y))
    # labels
    target = np.concatenate((np.zeros((X.shape[0],)), np.ones((Y.shape[0],))))

    shuffle = KFold(n_splits=n_folds, shuffle=True, random_state=seed)
    scores = cross_val_score(
        clf, data, target, cv=shuffle, scoring=metric, verbose=verbosity
    )

    return scores


def unbiased_mmd_squared(x, y):
    nx, ny = x.shape[0], y.shape[0]

    def f(a, b, diag=False):
        if diag:
            return torch.sum((a[None, ...] - b[:, None, :]) ** 2, dim=-1).reshape(-1)
        else:
            m, n = a.shape[0], b.shape[0]
            ix = torch.tril_indices(m, n, offset=-1)
            return torch.sum(
                (a[None, ...] - b[:, None, :]) ** 2, dim=-1, keepdim=False
            )[ix[0, :], ix[1, :]].reshape(-1)

    xx = f(x, x)
    xy = f(x, y, diag=True)
    yy = f(y, y)

    scale = torch.median(torch.sqrt(torch.cat((xx, xy, yy))))
    c = -0.5 / (scale**2)

    k = lambda a: torch.sum(torch.exp(c * a))

    kxx = k(xx) / (nx * (nx - 1))
    kxy = k(xy) / (nx * ny)
    kyy = k(yy) / (ny * (ny - 1))
    del xx, xy, yy

    mmd_square = 2 * (kxx + kyy - kxy)
    del kxx, kxy, kyy

    return mmd_square


def biased_mmd(x, y):
    nx, ny = x.shape[0], y.shape[0]

    def f(a, b):
        return torch.sum((a[None, ...] - b[:, None, :]) ** 2, dim=-1).reshape(-1)

    xx = f(x, x)
    xy = f(x, y)
    yy = f(y, y)

    scale = torch.median(torch.sqrt(torch.cat((xx, xy, yy))))
    c = -0.5 / (scale**2)

    k = lambda a: torch.sum(torch.exp(c * a))

    kxx = k(xx) / nx**2
    del xx
    kxy = k(xy) / (nx * ny)
    del xy
    kyy = k(yy) / ny**2
    del yy

    mmd_square = kxx - 2 * kxy + kyy
    del kxx, kxy, kyy

    return torch.sqrt(mmd_square)


def biased_mmd_hypothesis_test(x, y, alpha=0.05):
    assert x.shape[0] == y.shape[0]
    mmd_biased = biased_mmd(x, y).item()
    threshold = np.sqrt(2 / x.shape[0]) * (1 + np.sqrt(-2 * np.log(alpha)))

    return mmd_biased, threshold


def unbiased_mmd_squared_hypothesis_test(x, y, alpha=0.05):
    assert x.shape[0] == y.shape[0]
    mmd_square_unbiased = unbiased_mmd_squared(x, y).item()
    threshold = (4 / np.sqrt(x.shape[0])) * np.sqrt(-np.log(alpha))

    return mmd_square_unbiased, threshold


def _test():
    n = 2500
    x, y = torch.randn(n, 5), torch.randn(n, 5)
    print(unbiased_mmd_squared(x, y), biased_mmd(x, y))
    # mmd(x, y), sq_maximum_mean_discrepancy(tensor2numpy(x), tensor2numpy(y))
    # mmd_hypothesis_test(x, y, alpha=0.0001)
    # unbiased_mmd_squared_hypothesis_test(x, y)


def main():
    _test()


if __name__ == "__main__":
    main()



from __future__ import annotations

from abc import ABC, abstractmethod
from copy import deepcopy
from typing import Callable, Dict, List, Optional, Tuple, Type, Union

import numpy as np
import torch
from pyro.distributions import TransformedDistribution
from torch import Tensor, nn
from torch.distributions import Distribution
from torch.optim import ASGD, SGD, Adadelta, Adagrad, Adam, Adamax, AdamW, RMSprop
from torch.optim.lr_scheduler import (
    CosineAnnealingLR,
    CosineAnnealingWarmRestarts,
    CyclicLR,
    ExponentialLR,
    LambdaLR,
    StepLR,
)

from sbi.inference.potentials.base_potential import BasePotential
from sbi.samplers.vi.vi_utils import (
    filter_kwrags_for_func,
    make_object_deepcopy_compatible,
    move_all_tensor_to_device,
)
from sbi.types import Array
from sbi.utils import check_prior

_VI_method = {}


class DivergenceOptimizer(ABC):
    """This is a wrapper round a PyTorch optimizer and scheduler, which will be used to
    learn the variational distribution. It further contains some methods to evaluate
    convergence and to recover a valid state if something went wrong.

    This class contains one abstract method `_generate_loss_function` which must be
    implemented und determines the loss function used within variational inference."""

    def __init__(
        self,
        potential_fn: BasePotential,
        q: TransformedDistribution,
        prior: Optional[Distribution] = None,
        n_particles: int = 256,
        clip_value: float = 5.0,
        optimizer: Type[
            Union[SGD, Adam, Adadelta, RMSprop, Adagrad, Adamax, AdamW, ASGD]
        ] = Adam,
        scheduler: Type[
            Union[
                CosineAnnealingLR,
                ExponentialLR,
                CosineAnnealingWarmRestarts,
                CyclicLR,
                LambdaLR,
                StepLR,
            ]
        ] = ExponentialLR,
        eps: float = 1e-5,
        **kwargs,
    ):
        """This is a wrapper around a PyTorch optimizer which is used to minimize some
            loss for variational inference.

        Args:
            potential_fn: Potential function of the target i.e. the posterior density up
                to normalization constant.
            q: Variational distribution
            prior: Prior distribution, which will be used within the warmup, if given.
                Note that this will not affect the potential_fn, so make sure to have
                the same prior within it.
            n_particles: Number of samples used to estimate gradients.
            clip_value: Norm value on which gradients are clipped.
            optimizer: Base class for an pytorch optimizer. We support on of [SGD, Adam,
                Adadelta, RMSprop, Adagrad, Adamax, AdamW, ASGD].
            scheduler: Base class for an pytorch scheduler. We support on of
                [CosineAnnealingLR, ExponentialLR, CosineAnnealingWarmRestarts,
                CyclicLR, LambdaLR, StepLR]. Note that you may have to pass additional
                arguments of the scheduling method, this can be passed within the
                keyword arguments.
            eps: This value determines the sensitivity of the convergence checks.
            kwargs: All additional arguments associated with optimizer, scheduler such
                as learning_rates, gamma-values and so on. We refer to the documentation
                of each of the supported optimizers or schedulers for details.

        """

        self.potential_fn = potential_fn
        self.q = q
        check_prior(prior)
        self.prior = prior
        self.device = potential_fn.device
        self.to(self.device)

        self.n_particles = n_particles
        self.clip_value = clip_value
        self.learning_rate = kwargs.get("lr", 1e-3)
        self.retain_graph = kwargs.get("retain_graph", False)
        self._kwargs = kwargs

        # This prevents error that would stop optimization.
        self.q.set_default_validate_args(False)
        if prior is not None:
            self.prior.set_default_validate_args(False)  # type: ignore

        # Manage modules if present.
        if hasattr(self.q, "modules"):
            self.modules = nn.ModuleList(self.q.modules())
        else:
            self.modules = nn.ModuleList()
        self.modules.train()

        # Ensure that distribution has parameters and that these are on the right device
        if not hasattr(self.q, "parameters"):
            raise ValueError(
                "The variational distribution has no parameters to optimize."
            )
        self.to(self.device)

        # Keep a state to resolve invalid values
        self.state_dict = [para.data.clone() for para in self.q.parameters()]

        # Init optimizer and scheduler with correct arguments
        opt_kwargs = filter_kwrags_for_func(optimizer.__init__, kwargs)
        kwargs.pop("lr")  # This is just because CyclicLR Scheduler ...
        scheduler_kwargs = filter_kwrags_for_func(scheduler.__init__, kwargs)

        self._optimizer = optimizer(self.q.parameters(), **opt_kwargs)
        self._scheduler = scheduler(self._optimizer, **scheduler_kwargs)
        # Loss and summary
        self.eps = eps
        self.num_step = 0
        self.warm_up_was_done = False
        self.losses = torch.ones(2000)
        self.moving_average = torch.ones(2000)
        self.moving_std = torch.ones(2000)
        self.moving_slope = torch.ones(2000)

        # Hyperparameters to change adaptively
        self.HYPER_PARAMETERS = ["n_particles", "clip_value", "eps"]

    @abstractmethod
    def _loss(self, *args, **kwargs) -> Tuple[Tensor, Tensor]:
        """This should be overwritten based on the `_generate_loss_function`. And is
        the function that is evaluate to compute the loss."""
        raise NotImplementedError("Child classes implement the loss.")

    def to(self, device: str) -> None:
        """This will move all parameters to the correct device, both for likelihood and
        posterior"""
        self.device = device
        # These just ensure that everythink must be on the same device.
        move_all_tensor_to_device(self.potential_fn, self.device)
        move_all_tensor_to_device(self.q, self.device)
        if self.prior is not None:
            move_all_tensor_to_device(self.prior, self.device)

    def warm_up(self, num_steps: int, method: str = "prior") -> None:
        """This initializes q, either to follow the prior or the base distribution
        of the flow. This can increase training stability.

        Args:
            num_steps: Number of steps to train.
            method: Method for warmup.

        """
        if method == "prior" and self.prior is not None:
            inital_target = self.prior
        elif method == "identity":
            inital_target = torch.distributions.TransformedDistribution(
                self.q.base_dist, self.q.transforms[-1]
            )
        else:
            NotImplementedError(
                "The only implemented methods are `prior` and `identity`."
            )

        for _ in range(num_steps):
            self._optimizer.zero_grad()
            if self.q.has_rsample:
                samples = self.q.rsample((32,))
                logq = self.q.log_prob(samples)
                logp = inital_target.log_prob(samples)  # type: ignore
                loss = -torch.mean(logp - logq)
            else:
                samples = inital_target.sample((256,))  # type: ignore
                loss = -torch.mean(self.q.log_prob(samples))
            loss.backward(retain_graph=self.retain_graph)
            self._optimizer.step()
        # Denote that warmup was already done
        self.warm_up_was_done = True

    def update_state(self) -> None:
        """This updates the current state."""
        for state_para, para in zip(self.state_dict, self.q.parameters()):
            if torch.isfinite(para).all():
                state_para.data = para.data.clone()
            else:
                nn.init.uniform_(para, a=-0.5, b=0.5)

    def resolve_state(self, warm_up_rounds: int = 100) -> None:
        """In case the parameters become nan, this method will try to fix the current
        state

        Args:
            warm_up_rounds: Number of warm_up_round one should do after failure.
        """
        for state_para, para in zip(self.state_dict, self.q.parameters()):
            para.data = state_para.data.clone().to(para.device)
        self._optimizer.__init__(self.q.parameters(), self.learning_rate)
        self.warm_up(warm_up_rounds)

    def loss(self, x_o: Tensor) -> Tuple[Tensor, Tensor]:
        """Computes the loss function which is optimized.
        Args:
            x_o: Observed data as input.

        Returns:
            surrogated_loss : The loss that will be differentiated, hence this must be
                differentiable by PyTorch
            loss : This loss will be displayed and used to determine convergence. This
                should not  be differentiable.
        """
        return self._loss(x_o)

    def step(self, x_o: Tensor) -> None:
        """Performs one gradient step

        Args:
            x_o: Observation which is used.

        """
        self._optimizer.zero_grad()
        surrogate_loss, loss = self.loss(x_o.to(self.device))
        surrogate_loss.backward(retain_graph=self.retain_graph)
        if not torch.isfinite(surrogate_loss):
            self.resolve_state()
        nn.utils.clip_grad.clip_grad_norm_(self.q.parameters(), self.clip_value)
        self._optimizer.step()
        self._scheduler.step()
        self.update_loss_stats(loss.cpu())
        if (self.num_step % 50) == 0:
            self.update_state()

    def converged(self, considered_values: int = 50) -> bool:
        """Determines convergence based on a estimate of the slope of the loss
        function. If it is smaller than `eps` then this function will return true.

        Args:
            considered_values: Window over which we will average.

        Returns:
            bool: True if converged, else false.

        """
        if self.num_step < considered_values:
            return False
        else:
            m = self.moving_slope[
                self.num_step - considered_values : self.num_step
            ].mean()
            return abs(m).item() < self.eps

    def reset_loss_stats(self) -> None:
        """This will reset the loss statistics."""
        self.losses = torch.ones(2000)
        self.moving_average = torch.ones(2000)
        self.moving_std = torch.ones(2000)
        self.moving_slope = torch.ones(2000)
        self.num_step = 0

    def update_loss_stats(self, loss: Tensor, window: int = 20) -> None:
        """Updates current loss statistics of the optimizer

        Args:
            loss: New loss value
            window: Window size for running statistics
        """
        if self.num_step >= len(self.losses):
            self.losses = torch.hstack([self.losses, torch.zeros(2000)])
            self.moving_average = torch.hstack([self.moving_average, torch.zeros(2000)])
            self.moving_std = torch.hstack([self.moving_std, torch.zeros(2000)])
            self.moving_slope = torch.hstack([self.moving_slope, torch.zeros(2000)])

        if self.num_step == 0:
            self.losses[self.num_step] = loss
            self.moving_average[self.num_step] = loss
        elif not torch.isfinite(loss):
            loss = self.losses[self.num_step - 1]
            self.moving_average[self.num_step] = self.moving_average[self.num_step - 1]
            self.moving_std[self.num_step] = self.moving_std[self.num_step - 1]
            self.moving_slope[self.num_step] = self.moving_slope[self.num_step - 1]
        else:
            self.losses[self.num_step] = loss
            self.moving_average[self.num_step] = (
                self.moving_average[self.num_step - 1]
                + (loss - self.losses[max(self.num_step - window, 0)]) / window
            )
            self.moving_std[self.num_step] = abs(
                self.moving_std[self.num_step - 1]
                + (loss - self.losses[max(self.num_step - window, 0)]) / window
                - self.moving_std[self.num_step - 1] / (window - 1)
            )
            self.moving_slope[self.num_step] = (
                loss - self.moving_average[max(self.num_step - window, 0)]
            ) / window

        self.num_step += 1

    def get_loss_stats(self) -> Tuple[Array, Array]:
        """Returns current loss statistics"""
        return (
            self.moving_average[self.num_step - 1],
            self.moving_std[self.num_step - 1],
        )

    def update(self, kwargs: Dict):
        """Updates the hyperparameters and scheduler/optimizer kwargs"""
        # print(kwargs)
        paras = self.__dict__
        for key, val in kwargs.items():
            if key == "retain_graph":
                self.retain_graph = val
            if key == "learning_rate":
                self.learning_rate = val
            if key in self.HYPER_PARAMETERS:
                paras[key] = val

            if key == "self":
                posterior = kwargs[key]
                self.q = posterior.q
                self.potential_fn = posterior.potential_fn
                self.prior = posterior._prior

        if "self" in kwargs:
            kwargs.pop("self")

        if "optimizer" in kwargs:
            opt_type = kwargs.pop("optimizer")
        else:
            opt_type = type(self._optimizer)
        if "scheduler" in kwargs:
            sched_type = kwargs.pop("scheduler")
        else:
            sched_type = type(self._scheduler)

        kwargs["lr"] = self.learning_rate
        opt_kwargs = filter_kwrags_for_func(opt_type.__init__, kwargs)
        kwargs.pop("lr")  # This is just because CyclicLR Scheduler ...
        scheduler_kwargs = filter_kwrags_for_func(sched_type.__init__, kwargs)

        self._optimizer = opt_type(self.q.parameters(), **opt_kwargs)
        self._scheduler = sched_type(self._optimizer, **scheduler_kwargs)


def register_VI_method(
    name: Optional[str] = None,
) -> Callable:
    """Registers a new VI method, by adding a new Divergence Optimizer class.

    Args:
        cls: Class to add
        name: Associated name
    """

    def _register(cls: "DivergenceOptimizer") -> "DivergenceOptimizer":
        if name is None:
            cls_name = cls.__class__.__name__
        else:
            cls_name = name
        if cls_name in _VI_method:
            raise ValueError(f"The VI method {cls_name} is already registered")
        else:
            _VI_method[cls_name] = cls
        return cls

    return _register


def get_VI_method(name: str):
    """Returns a specific DivergenceOptimizer using the specified VI method.

    Args:
        name: The name of the method

    Returns:
        DivergenceOptimizer: An divergence optimizer.

    """
    return _VI_method[name]


def get_default_VI_method() -> List[str]:
    return list(_VI_method.keys())


@register_VI_method(name="rKL")
class ElboOptimizer(DivergenceOptimizer):
    r"""This will learn the variational posterior by minimizing the reverse KL
    divergence between q and p i.e. $D_KL(q(\theta)||p(\theta|x_o))$, by maximizing the
    ELBO (Evidence Lower Bound).
    """

    def __init__(self, *args, stick_the_landing: bool = False, **kwargs):
        """See `DivergenceOptimizer` for all the base arguments.

        Args:
            stick_the_landing: This will reduce the variance of the estimator,especially
                near convergence. Yet for normalizing flows this adds additional cost as
                it requires to evaluate the inverse pass which often can be avoided
                through caching. See [1] for details.

        References:
            [1] Sticking the Landing: Simple, Lower-Variance Gradient Estimators for
                Variational Inference, Geoffrey Roeder, Yuhuai Wu, David Duvenaud, 2017,
                https://arxiv.org/abs/1703.09194.

        """
        super().__init__(*args, **kwargs)

        self.stick_the_landing = stick_the_landing
        make_object_deepcopy_compatible(self.q)
        self._surrogate_q = deepcopy(self.q)

        self.eps = 1e-5
        self.HYPER_PARAMETERS += ["stick_the_landing"]

    def _loss(self, xo: Tensor) -> Tuple[Tensor, Tensor]:
        if self.q.has_rsample:
            return self.loss_rsample(xo)
        else:
            raise NotImplementedError(
                "Currently only reparameterizable distributions are supported."
            )

    def loss_rsample(self, x_o: Tensor) -> Tuple[Tensor, Tensor]:
        """Computes the ELBO"""
        elbo_particles = self.generate_elbo_particles(x_o)
        loss = -elbo_particles.mean()
        return loss, loss.clone().detach()

    def generate_elbo_particles(
        self, x_o: Tensor, num_samples: Optional[int] = None
    ) -> Tensor:
        """Generates individual ELBO particles i.e. logp(theta, x_o) - logq(theta)."""
        if num_samples is None:
            num_samples = self.n_particles
        samples = self.q.rsample((num_samples,))
        if self.stick_the_landing:
            self.update_surrogate_q()
            log_q = self._surrogate_q.log_prob(samples)
        else:
            log_q = self.q.log_prob(samples)
        self.potential_fn.x_o = x_o
        log_potential = self.potential_fn(samples)
        elbo = log_potential - log_q
        return elbo

    def update_surrogate_q(self) -> None:
        """Updates the surrogate with new parameters."""
        for param, param_surro in zip(
            self.q.parameters(), self._surrogate_q.parameters()
        ):
            param_surro.data = param.data
            param_surro.requires_grad = False


@register_VI_method(name="IW")
class IWElboOptimizer(ElboOptimizer):
    r"""This learns the variational posterior by minimizing the importance weighted
    ELBO, which is an tighter bound to the evidence but also promotes a support covering
    behaviour.

    NOTE: You may want to turn on `stick_the_landing` here as this loss leads to
    gradient
    estimates with vanishing signal to noise ratio. This is relevant for large K,
    especially K > n_particles.

    NOTE: Technically this does not minimize a valid divergence between q and p, yet it
    does optimizer q as proposal for sampling importance resampling.

    References:
        Importance Weighted Autoencoders, Yuri Burda, Roger Grosse, Ruslan
        Salakhutdinov, 2016, https://arxiv.org/abs/1509.00519.
    """

    def __init__(self, *args, K: int = 8, dreg: bool = False, **kwargs):
        """See `DivergenceOptimizer` for all the base arguments.



        Args:
            K: Number of samples used within estimating a single gradient. In total
                n_particles x K samples are used.
            dreg: Doubly reparmeterized gradient estimator as proposed in [1]. It is
                based on the `stick_the_landing` already present but leads to an
                unbiased estimate on this objective in contrast.

        References:
            [1] _Doubly Reparameterized Gradient Estimators for Monte Carlo Objectives_,
                George Tucker, Dieterich Lawson, Shixiang Gu, Chris J. Maddison, 2018,
                https://arxiv.org/abs/1810.04152.


        """
        super().__init__(*args, **kwargs)
        self.K = K
        self.loss_name = "iwelbo"
        self.eps = 1e-7
        self.dreg = dreg
        self.HYPER_PARAMETERS += ["K", "dreg"]
        if dreg:
            self.stick_the_landing = True

    def loss_rsample(self, x_o: Tensor) -> Tuple[Tensor, Tensor]:
        """Computes the IWELBO loss.

        Args:
            x_o: Observation

        Returns:
            surrogate_loss: Loss to differentiate.
            loss: Loss to diplay.

        """
        elbo_particles = self.generate_elbo_particles(x_o, self.n_particles * self.K)
        elbo_particles = elbo_particles.reshape(self.n_particles, self.K)
        weights = self.get_importance_weight(elbo_particles.clone().detach())
        surrogate_loss = -(weights * elbo_particles).sum(-1).mean(0)
        loss = -torch.mean(torch.exp(elbo_particles) + 1e-20, -1).log().mean()
        return surrogate_loss, loss.clone().detach()

    def get_importance_weight(self, elbo_particles: Tensor) -> Tensor:
        """Generates importance weights in a numerically stable fashion.

        Args:
            elbo_particles: ELements contain log potential - log q.

        Returns:
            Tensor: Normalized importance weights.

        """
        logweights = elbo_particles
        normalized_weights = torch.exp(
            logweights - torch.logsumexp(logweights, -1).unsqueeze(-1)
        )
        if self.dreg:
            normalized_weights = normalized_weights**2
        return normalized_weights


@register_VI_method(name="fKL")
class ForwardKLOptimizer(DivergenceOptimizer):
    """This learns the variational posterior by minimizing the forward KL divergence
    i.e. $D_KL(p(\theta|x_o)|| q(\theta))$. The typically necessary samples from the
    posterior are not required by using importance sampling techniques.

    NOTE: Whereas in the rKL and IWELBO n_particles mainly reduces the variance, in this
    case it also decrease the bias. Using n_particles=1 is not usefull.

    References:
        [1] _Variational Refinement for Importance Sampling Using the Forward
            Kullback-Leibler Divergence_, Ghassen Jerfel, Serena Wang, Clara Fannjiang,
            Katherine A. Heller, Yian Ma, Michael I. Jordan, 2021,
            https://arxiv.org/abs/2106.15980.
    """

    def __init__(
        self,
        *args,
        proposal: str = "q",
        weight_transform: Callable = lambda x: x,
        **kwargs,
    ):
        """See `DivergenceOptimizer` for base arguments.

        Args:
            proposal: The proposal used we currently support only [`q`]
            weight_transform: This is a transformation that is applied to the importance
                weight $w(\theta) = p(\theta, x_o)/q(\theta)$. For example it may be
                usefull to clamp the weights by a certain rule to reduce variance.


        """
        super().__init__(*args, **kwargs)

        self.proposal = proposal
        self.weight_transform = weight_transform
        self._loss_name = "forward_kl"
        self.HYPER_PARAMETERS += ["weight_transform", "proposal"]
        self.eps = 1e-5

    def _loss(self, xo: Tensor) -> Tuple[Tensor, Tensor]:
        if self.q.has_rsample:
            return self._loss_q_proposal(xo)
        else:
            raise NotImplementedError("Unknown loss.")

    def _loss_q_proposal(self, x_o: Tensor) -> Tuple[Tensor, Tensor]:
        """This gives an importance sampling estimate of the forward KL divergence.

        Args:
            x_o: Observation.

        Returns:
            Tuple[Tensor, Tensor]: Surrogate loss to differentiate and to display.

        """
        samples = self.q.sample((self.n_particles,))
        if hasattr(self.q, "clear_cache"):
            self.q.clear_cache()
        logq = self.q.log_prob(samples)
        self.potential_fn.x_o = x_o
        logp = self.potential_fn(samples)
        with torch.no_grad():
            logweights = logp - logq
            weights = self.weight_transform(logweights.exp())
            weights /= weights.sum()

        surrogate = -torch.sum(weights * logq)
        loss = torch.sum(weights * (logp - logq))
        return surrogate, loss.detach()


@register_VI_method(name="alpha")
class RenyiDivergenceOptimizer(ElboOptimizer):
    r"""This learns the variational posterior by minimizing Renyi alpha divergences. For
    alpha=0 we obtain this is equivalent to the IWELBO with n_particles=1 and
    K=n_particles. For alpha=1 this is equivalent to the `ElboOptimizer`.

    NOTE: For alpha < 1 the divergence is more mass covering, for alpha > 1 the
    divergence is more mode seeking.

    NOTE: For small alpha, you i.e. alpha=0.1 you may require to
    turn on stick_the_landing or dreg.

    References:
        [1] _Rényi Divergence Variational Inference_, Yingzhen Li, Richard E. Turner,
            2016,https://arxiv.org/abs/1602.02311.
    """

    def __init__(
        self,
        *args,
        alpha: float = 0.5,
        unbiased: bool = False,
        dreg: bool = False,
        **kwargs,
    ):
        """See `ElboOptimizer` for the base arguments.

        Args:
            alpha: Determines the alpha divergence to which is used. For alpha < 1 the
                divergence is more mass covering, for alpha > 1 the divergence is more
                mode seeking.
            unbiased: We use the biased bound as proposed in [1], but one can also use
                an unbiased one.
            dreg: Doubly reparmeterized gradient estimator as proposed in [2]. It is
                based on the `reduced_variance` already present, but leads to an
                unbiased estimate on this objective in contrast.

        Reference:
            [1] _Rényi Divergence Variational Inference_, Yingzhen Li, Richard E.
                Turner, 2016,https://arxiv.org/abs/1602.02311.
            [2] _Doubly Reparameterized Gradient Estimators for Monte Carlo Objectives_,
                George Tucker, Dieterich Lawson, Shixiang Gu, Chris J. Maddison, 2018,
                https://arxiv.org/abs/1810.04152.

        """
        self.alpha = alpha
        self.unbiased = unbiased
        super().__init__(*args, **kwargs)
        self.HYPER_PARAMETERS += ["alpha", "unbiased", "dreg"]
        self.eps = 1e-5
        self.dreg = dreg
        if dreg:
            self.stick_the_landing = True

    def _loss(self, xo: Tensor) -> Tuple[Tensor, Tensor]:
        assert isinstance(self.alpha, float)
        if self.q.has_rsample:
            if not self.unbiased:
                return self.loss_alpha(xo)
            else:
                return self.loss_alpha_unbiased(xo)
        else:
            raise NotImplementedError(
                "Currently we only support reparameterizable distributions"
            )

    def loss_alpha_unbiased(self, x_o: Tensor) -> Tuple[Tensor, Tensor]:
        """Unbiased estimate of a surrogate RVB"""
        elbo_particles = self.generate_elbo_particles(x_o) * (1 - self.alpha)
        _, mean_log_weights = self.get_importance_weight(elbo_particles)
        surrogate_loss = -torch.exp(mean_log_weights)
        loss = -mean_log_weights / (1 - self.alpha)
        return surrogate_loss, loss.clone().detach()

    def loss_alpha(self, x_o: Tensor) -> Tuple[Tensor, Tensor]:
        """Renjy variation bound (RVB)."""
        elbo_particles = self.generate_elbo_particles(x_o)
        weights, mean_log_weights = self.get_importance_weight(
            elbo_particles.clone().detach()
        )
        surrogate_loss = -torch.mean(weights * elbo_particles)
        loss = -mean_log_weights / (1 - self.alpha)
        return surrogate_loss, loss

    def get_importance_weight(self, elbo_particles: Tensor) -> Tuple[Tensor, Tensor]:
        """Computes the importance weights for the gradients"""
        logweights = (1 - self.alpha) * elbo_particles
        mean_log_weights = torch.logsumexp(logweights, 0) - np.log(self.n_particles)
        normed_logweights = logweights - mean_log_weights
        weights = normed_logweights.exp()
        if self.dreg:
            weights = weights**2
        return weights, mean_log_weights



# This file is part of sbi, a toolkit for simulation-based inference. sbi is licensed
# under the Affero General Public License v3, see <https://www.gnu.org/licenses/>.


from typing import Callable, Optional, Union

import torch
from torch import Tensor
from torch.distributions import Distribution

import sbi.utils as utils
from sbi.inference.snpe.snpe_base import PosteriorEstimator
from sbi.types import TensorboardSummaryWriter
from sbi.utils import del_entries


class SNPE_B(PosteriorEstimator):
    def __init__(
        self,
        prior: Optional[Distribution] = None,
        density_estimator: Union[str, Callable] = "maf",
        device: str = "cpu",
        logging_level: Union[int, str] = "WARNING",
        summary_writer: Optional[TensorboardSummaryWriter] = None,
        show_progress_bars: bool = True,
    ):
        r"""SNPE-B [1]. CURRENTLY NOT IMPLEMENTED.

        [1] _Flexible statistical inference for mechanistic models of neural dynamics_,
            Lueckmann, Gonçalves et al., NeurIPS 2017, https://arxiv.org/abs/1711.01861.

        See docstring of `PosteriorEstimator` class for all other arguments.
        """

        raise NotImplementedError(
            "SNPE-B is not yet implemented in the sbi package, see issue #199."
        )

        kwargs = del_entries(locals(), entries=("self", "__class__"))
        super().__init__(**kwargs)

    def _log_prob_proposal_posterior(
        self, theta: Tensor, x: Tensor, masks: Tensor
    ) -> Tensor:
        """
        Return importance-weighted log probability (Lueckmann, Goncalves et al., 2017).

        Args:
            theta: Batch of parameters θ.
            x: Batch of data.
            masks: Whether to retrain with prior loss (for each prior sample).

        Returns:
            Log probability of proposal posterior.
        """

        raise NotImplementedError

        batch_size = theta.shape[0]

        # Evaluate posterior.
        log_prob_posterior = self._posterior.net.log_prob(theta, x)
        log_prob_posterior = log_prob_posterior.reshape(batch_size)
        utils.assert_all_finite(log_prob_posterior, "posterior eval")

        # Evaluate prior.
        log_prob_prior = self._prior.log_prob(theta).reshape(batch_size)
        utils.assert_all_finite(log_prob_prior, "prior eval.")

        # Evaluate proposal.
        log_prob_proposal = self._model_bank[-1].net.log_prob(theta, x)
        utils.assert_all_finite(log_prob_proposal, "proposal posterior eval")

        # Compute log prob with importance weights.
        log_prob = torch.exp(log_prob_prior - log_prob_proposal) * log_prob_posterior

        return log_prob



# This file is part of sbi, a toolkit for simulation-based inference. sbi is licensed
# under the Affero General Public License v3, see <https://www.gnu.org/licenses/>.

import logging
from copy import deepcopy
from typing import Any, Callable, Optional, Tuple, Union
from warnings import warn

import torch
from pyknos.nflows.nn import nets
from torch import Tensor, nn, optim, relu
from torch.nn import MSELoss
from torch.nn.utils.clip_grad import clip_grad_norm_
from torch.utils import data
from torch.utils.data.sampler import SubsetRandomSampler

from sbi.utils.sbiutils import handle_invalid_x, standardizing_net


class Destandardize(nn.Module):
    def __init__(self, mean: Union[Tensor, float], std: Union[Tensor, float]):
        super(Destandardize, self).__init__()
        mean, std = map(torch.as_tensor, (mean, std))
        self.mean = mean
        self.std = std
        self.register_buffer("_mean", mean)
        self.register_buffer("_std", std)

    def forward(self, tensor):
        return tensor * self._std + self._mean


def destandardizing_net(batch_t: Tensor, min_std: float = 1e-7) -> nn.Module:
    """Net that de-standardizes the output so the NN can learn the standardized target.

    Args:
        batch_t: Batched tensor from which mean and std deviation (across
            first dimension) are computed.
        min_std:  Minimum value of the standard deviation to use when z-scoring to
            avoid division by zero.

    Returns:
        Neural network module for z-scoring
    """

    is_valid_t, *_ = handle_invalid_x(batch_t, True)

    t_mean = torch.mean(batch_t[is_valid_t], dim=0)
    if len(batch_t > 1):
        t_std = torch.std(batch_t[is_valid_t], dim=0)
        t_std[t_std < min_std] = min_std
    else:
        t_std = 1
        logging.warning(
            """Using a one-dimensional batch will instantiate a Standardize transform
            with (mean, std) parameters which are not representative of the data. We
            allow this behavior because you might be loading a pre-trained. If this is
            not the case, please be sure to use a larger batch."""
        )

    return Destandardize(t_mean, t_std)


def build_input_output_layer(
    batch_theta: Tensor,
    batch_property: Tensor,
    z_score_theta: bool = True,
    z_score_property: bool = True,
    embedding_net_theta: nn.Module = nn.Identity(),
) -> Tuple[nn.Module, nn.Module]:
    r"""Builds input layer for the `ActiveSubspace` that optionally z-scores.

    The regression network used in the `ActiveSubspace` will receive batches of
    $\theta$s and properties.

    Args:
        batch_theta: Batch of $\theta$s, used to infer dimensionality and (optional)
            z-scoring.
        batch_property: Batch of properties, used for (optional) z-scoring.
        z_score_theta: Whether to z-score $\theta$s passing into the network.
        z_score_property: Whether to z-score properties passing into the network.
        embedding_net_theta: Optional embedding network for $\theta$s.

    Returns:
        Input layer that optionally z-scores.
    """

    if z_score_theta:
        input_layer = nn.Sequential(standardizing_net(batch_theta), embedding_net_theta)
    else:
        input_layer = embedding_net_theta

    if z_score_property:
        output_layer = destandardizing_net(batch_property)
    else:
        output_layer = nn.Identity()

    return input_layer, output_layer


class ActiveSubspace:
    def __init__(self, posterior: Any):
        """Identify the active subspace for sensitivity analyses.

        - Introduction to active subspaces: Constantine et al. 2015.
        - Application to analyse the sensitivity in neuroscience models:
            Deistler et al. 2021, in preparation.

        Args:
            posterior: Posterior distribution obtained with `SNPE`, `SNLE`, or `SNRE`.
                Needs to have a `.sample()` method. If we want to analyse the
                sensitivity of the posterior probability, it also must have a
                `.potential()` method.
        """
        self._posterior = posterior
        self._regression_net = None
        self._theta = None
        self._emergent_property = None
        self._device = posterior._device
        self._validation_log_probs = []

    def add_property(
        self,
        theta: Tensor,
        emergent_property: Tensor,
        model: Union[str, Callable] = "resnet",
        hidden_features: int = 100,
        num_blocks: int = 2,
        dropout_probability: float = 0.5,
        z_score_theta: bool = True,
        z_score_property: bool = True,
        embedding_net: nn.Module = nn.Identity(),
    ) -> "ActiveSubspace":
        r"""Add a property whose sensitivity is to be analysed.

        To analyse the sensitivity of an emergent property, we train a neural network
        to predict the property from the parameter set $\theta$. The hyperparameters of
        this neural network also have to be specified here.

        Args:
            theta: Parameter sets $\theta$ sampled from the posterior.
            emergent_property: Tensor containing the values of the property given each
                parameter set $\theta$.
            model: Neural network used to distinguish valid from bad samples. If it is
                a string, use a pre-configured network of the provided type (either
                mlp or resnet). Alternatively, a function that builds a custom
                neural network can be provided. The function will be called with the
                first batch of parameters (theta,), which can thus be used for shape
                inference and potentially for z-scoring. It needs to return a PyTorch
                `nn.Module` implementing the classifier.
            hidden_features: Number of hidden units of the classifier if `model` is a
                string.
            num_blocks: Number of hidden layers of the classifier if `model` is a
                string.
            dropout_probability: Dropout probability of the classifier if `model` is
                `resnet`.
            z_score_theta: Whether to z-score the parameters $\theta$ used to train the
                classifier.
            z_score_property: Whether to z-score the property used to train the
                classifier.
            embedding_net: Neural network used to encode the parameters before they are
                passed to the classifier.

        Returns:
            `ActiveSubspace` to make the call chainable.
        """
        assert emergent_property.shape == (
            theta.shape[0],
            1,
        ), "The `emergent_property` must have shape (N, 1)."

        self._theta = theta
        self._emergent_property = emergent_property

        def build_resnet(theta):
            classifier = nets.ResidualNet(
                in_features=theta.shape[1],
                out_features=1,
                hidden_features=hidden_features,
                context_features=None,
                num_blocks=num_blocks,
                activation=relu,
                dropout_probability=dropout_probability,
                use_batch_norm=True,
            )
            input_layer, output_layer = build_input_output_layer(
                theta,
                emergent_property,
                z_score_theta,
                z_score_property,
                embedding_net,
            )
            classifier = nn.Sequential(input_layer, classifier, output_layer)
            return classifier

        def build_mlp(theta):
            classifier = nn.Sequential(
                nn.Linear(theta.shape[1], hidden_features),
                nn.BatchNorm1d(hidden_features),
                nn.ReLU(),
                nn.Linear(hidden_features, hidden_features),
                nn.BatchNorm1d(hidden_features),
                nn.ReLU(),
                nn.Linear(hidden_features, 1),
            )
            input_layer, output_layer = build_input_output_layer(
                theta,
                emergent_property,
                z_score_theta,
                z_score_property,
                embedding_net,
            )
            classifier = nn.Sequential(input_layer, classifier, output_layer)
            return classifier

        if isinstance(model, str):
            if model == "resnet":
                self._build_nn = build_resnet
            elif model == "mlp":
                self._build_nn = build_mlp
            else:
                raise NameError
        else:
            self._build_nn = model

        return self

    def train(
        self,
        training_batch_size: int = 50,
        learning_rate: float = 5e-4,
        validation_fraction: float = 0.1,
        stop_after_epochs: int = 20,
        max_num_epochs: int = 2**31 - 1,
        clip_max_norm: Optional[float] = 5.0,
    ) -> nn.Module:
        r"""Train a regression network to predict the specified property from $\theta$.

        Args:
            training_batch_size: Training batch size.
            learning_rate: Learning rate for Adam optimizer.
            validation_fraction: The fraction of data to use for validation.
            stop_after_epochs: The number of epochs to wait for improvement on the
                validation set before terminating training.
            max_num_epochs: Maximum number of epochs to run. If reached, we stop
                training even when the validation loss is still decreasing. Otherwise,
                we train until validation loss increases (see also `stop_after_epochs`).
            clip_max_norm: Value at which to clip the total gradient norm in order to
                prevent exploding gradients. Use `None` for no clipping.
        """

        assert (
            self._theta is not None and self._emergent_property is not None
        ), "You must call .add_property() first."

        # Get indices for permutation of the data.
        num_examples = len(self._theta)
        permuted_indices = torch.randperm(num_examples)
        num_training_examples = int((1 - validation_fraction) * num_examples)
        num_validation_examples = num_examples - num_training_examples
        train_indices, val_indices = (
            permuted_indices[:num_training_examples],
            permuted_indices[num_training_examples:],
        )

        # Dataset is shared for training and validation loaders.
        dataset = data.TensorDataset(self._theta, self._emergent_property)

        # Create neural_net and validation loaders using a subset sampler.
        train_loader = data.DataLoader(
            dataset,
            batch_size=training_batch_size,
            drop_last=True,
            sampler=SubsetRandomSampler(train_indices.tolist()),
        )
        val_loader = data.DataLoader(
            dataset,
            batch_size=min(training_batch_size, num_examples - num_training_examples),
            shuffle=False,
            drop_last=True,
            sampler=SubsetRandomSampler(val_indices.tolist()),
        )

        if self._regression_net is None:
            self._regression_net = self._build_nn(self._theta[train_indices]).to(
                self._device
            )

        optimizer = optim.Adam(
            list(self._regression_net.parameters()),
            lr=learning_rate,
        )
        max_num_epochs = 2**31 - 1 if max_num_epochs is None else max_num_epochs

        # criterion / loss
        criterion = MSELoss()

        epoch, self._val_log_prob = 0, float("-Inf")
        while epoch <= max_num_epochs and not self._converged(epoch, stop_after_epochs):
            self._regression_net.train()
            for parameters, observations in train_loader:
                optimizer.zero_grad()
                outputs = self._regression_net(parameters.to(self._device))
                loss = criterion(outputs, observations.to(self._device))
                loss.backward()
                if clip_max_norm is not None:
                    clip_grad_norm_(
                        self._regression_net.parameters(),
                        max_norm=clip_max_norm,
                    )
                optimizer.step()

            epoch += 1

            # calculate validation performance
            self._regression_net.eval()

            val_loss = 0.0
            with torch.no_grad():
                for parameters, observations in val_loader:
                    outputs = self._regression_net(parameters.to(self._device))
                    loss = criterion(outputs, observations.to(self._device))
                    val_loss += loss.item()
            self._val_log_prob = -val_loss / num_validation_examples
            self._validation_log_probs.append(self._val_log_prob)

            print("\r", "Training neural network. Epochs trained: ", epoch, end="")

        return deepcopy(self._regression_net)

    def _converged(self, epoch: int, stop_after_epochs: int) -> bool:
        r"""Return whether the training converged yet and save best model state so far.

        Checks for improvement in validation performance over previous epochs.

        Args:
            epoch: Current epoch in training.
            stop_after_epochs: How many fruitless epochs to let pass before stopping.
        Returns:
            Whether the training has stopped improving, i.e. has converged.
        """
        converged = False

        assert self._regression_net is not None
        posterior_nn = self._regression_net

        # (Re)-start the epoch count with the first epoch or any improvement.
        if epoch == 0 or self._val_log_prob > self._best_val_log_prob:
            self._best_val_log_prob = self._val_log_prob
            self._epochs_since_last_improvement = 0
            self._best_model_state_dict = deepcopy(posterior_nn.state_dict())
        else:
            self._epochs_since_last_improvement += 1

        # If no validation improvement over many epochs, stop training.
        if self._epochs_since_last_improvement > stop_after_epochs - 1:
            posterior_nn.load_state_dict(self._best_model_state_dict)
            converged = True

        return converged

    def find_directions(
        self,
        posterior_log_prob_as_property: bool = False,
        norm_gradients_to_prior: bool = True,
        num_monte_carlo_samples: int = 1000,
    ) -> Tuple[Tensor, Tensor]:
        r"""Return eigenvectors and values corresponding to directions of sensitivity.

        The directions of sensitivity are the directions along which a specific
        property changes in the fastest way. They will have the largest eigenvalues.

        This computes the matrix:
        $\mathbf{M} = \mathbb{E}_{p(\theta|x_o)}[\nabla_{\theta} f(\theta)^T
        \nabla_{\theta}
        f(\theta)]$
        where $f(\cdot)$ is the trained regression network. The expected value is
        approximated with a Monte-Carlo mean. Next, do an eigenvalue
        decomposition of the matrix $\mathbf{M}$:

        $\mathbf{M} = \mathbf{Q} \mathbf{\Lambda} \mathbf{Q}^{-1}$

        We then return the eigenvectors and eigenvalues found by this decomposition.
        Eigenvectors with large eigenvalues are directions along which the property is
        sensitive to changes in the parameters $\theta$ (`active` directions).
        Increases along these directions will increase the value of the property.

        Args:
            posterior_log_prob_as_property: Whether to use the posterior
                log-probability the key property whose sensitivity is analysed. If
                `False`, one must have specified an emergent property and trained a
                regression network using `.add_property().train()`. If `True`,
                any previously specified property is ignored.
            norm_gradients_to_prior: Whether to normalize each entry of the gradient
                by the standard deviation of the prior in each dimension. If set to
                `False`, the directions with the strongest eigenvalues might correspond
                to directions in which the prior is broad.
            num_monte_carlo_samples: Number of Monte Carlo samples that the average is
                based on. A larger value will make the results more accurate while
                requiring more compute time.

        Returns:
            Eigenvectors and corresponding eigenvalues. They are sorted in ascending
            order. The column `eigenvectors[:, j]` is the eigenvector corresponding to
            the `j`-th eigenvalue.
        """

        self._gradients_are_normed = norm_gradients_to_prior

        if self._emergent_property is None and not posterior_log_prob_as_property:
            raise ValueError(
                "You have not yet passed an emergent property whose "
                "sensitivity you would like to analyse. Please use "
                "`.add_emergent_property().train()` to do so. If you want "
                "to use all features that had also been used to infer the "
                "posterior distribution (i.e. you want to analyse the "
                "sensitivity of the posterior probability), use: "
                "`.find_active(posterior_log_prob_as_property=True)`."
            )
        if self._emergent_property is not None and posterior_log_prob_as_property:
            warn(
                "You specified a property with `.add_property()`, but also set "
                "`posterior_log_prob_as_property=True`. The specified property will "
                "be ignored."
            )

        thetas = self._posterior.sample((num_monte_carlo_samples,))

        thetas.requires_grad = True

        if posterior_log_prob_as_property:
            predictions = self._posterior.potential(thetas, track_gradients=True)
        else:
            assert (
                self._regression_net is not None
            ), "self._regression_net is None, you must call `.train()` first."
            predictions = self._regression_net.forward(thetas)
        loss = predictions.mean()
        loss.backward()
        gradients = torch.squeeze(thetas.grad)
        if norm_gradients_to_prior:
            if hasattr(self._posterior.prior, "stddev") and hasattr(
                self._posterior.prior, "mean"
            ):
                self._prior_mean = self._posterior.prior.mean
                self._prior_scale = self._posterior.prior.stddev
            else:
                prior_samples = self._posterior.prior.sample((10000,))
                self._prior_scale = torch.std(prior_samples, dim=0)
                self._prior_mean = torch.mean(prior_samples, dim=0)
            gradients *= self._prior_scale
        outer_products = torch.einsum("bi,bj->bij", (gradients, gradients))
        average_outer_product = outer_products.mean(dim=0)

        eigen_values, eigen_vectors = torch.linalg.eigh(average_outer_product, UPLO="U")

        # Identify the direction of the eigenvectors. Above, we have computed an outer
        # product m*mT=A. Note that the same matrix A can be constructed with the
        # negative vector (-m)(-mT)=A. Thus, when performing an eigen-decomposition of
        # A, we can not determine if the eigenvector was -m or m. We solve this issue
        # below. We use that the average gradient m should be obtained by a mean over
        # the eigenvectors (weighted by the eigenvalues).
        av_gradient = torch.mean(gradients, dim=0)
        av_gradient = av_gradient / torch.norm(av_gradient)
        av_eigenvec = torch.mean(eigen_vectors * eigen_values, dim=1)
        av_eigenvec = av_eigenvec / torch.norm(av_eigenvec)

        # Invert if the negative eigenvectors are closer to the average gradient.
        if (torch.mean((av_eigenvec - av_gradient) ** 2)) > (
            torch.mean((-av_eigenvec - av_gradient) ** 2)
        ):
            eigen_vectors = -eigen_vectors

        self._eigen_vectors = eigen_vectors

        return eigen_values, eigen_vectors

    def project(self, theta: Tensor, num_dimensions: int) -> Tensor:
        r"""Return $\theta$ that were projected into the subspace.

        To identify the dimensionality of the active subspace `num_dimensions`,
        Constantine et al. 2015 suggest to look at gaps in the eigenvalue spectrum.

        Performs a linear projection. Also takes care of normalizing the data. The mean
        and standard deviation used for normalizing are the same as used to compute the
        eigenvectors and eigenvalues (mean and std of prior).

        Args:
            theta: Parameter sets to be projected.
            num_dimensions: Dimensionality of the subspace into which to project.

        Returns:
            Projected parameters of shape `(theta.shape[0], num_dimensions)`.
        """
        theta = theta.to(self._device)
        if self._gradients_are_normed:
            theta = (theta - self._prior_mean) / self._prior_scale

        projection_mat = self._eigen_vectors[:, -num_dimensions:]
        projected_theta = torch.mm(theta, projection_mat)

        return projected_theta



from copy import deepcopy
from math import floor
from typing import Callable, Optional, Tuple, Union

import torch
import torch.nn.functional as F
from pyknos.nflows.nn import nets
from torch import Tensor, nn, optim, relu
from torch.distributions import Distribution
from torch.nn.utils.clip_grad import clip_grad_norm_
from torch.utils import data
from torch.utils.data.sampler import SubsetRandomSampler, WeightedRandomSampler
from tqdm.auto import tqdm

from sbi.types import Shape
from sbi.utils.sbiutils import (
    get_simulations_since_round,
    handle_invalid_x,
    standardizing_net,
    z_score_parser,
)
from sbi.utils.torchutils import ensure_theta_batched
from sbi.utils.user_input_checks import validate_theta_and_x


def build_input_layer(
    batch_theta: Tensor,
    z_score_theta: Optional[str] = "independent",
    embedding_net_theta: nn.Module = nn.Identity(),
) -> nn.Module:
    r"""Builds input layer for the `RestrictionEstimator` with option to z-score.

    The classifier used in the `RestrictionEstimator` will receive batches of $\theta$s.

    Args:
        batch_theta: Batch of $\theta$s, used to infer dimensionality and (optional)
            z-scoring.
        z_score_theta: Whether to z-score parameters $\theta$ before passing them into
            the network, can take one of the following:
            - `none`, or None: do not z-score.
            - `independent`: z-score each dimension independently.
            - `structured`: treat dimensions as related, therefore compute mean and std
            over the entire batch, instead of per-dimension. Should be used when each
            sample is, for example, a time series or an image.
        embedding_net_theta: Optional embedding network for $\theta$s.

    Returns:
        Input layer with optional embedding net and z-scoring.
    """
    z_score_theta_bool, structured_theta = z_score_parser(z_score_theta)
    if z_score_theta_bool:
        input_layer = nn.Sequential(
            standardizing_net(batch_theta, structured_theta), embedding_net_theta
        )
    else:
        input_layer = embedding_net_theta

    return input_layer


def build_classifier(
    model: Union[str, Callable] = "resnet",
    hidden_features: int = 100,
    num_blocks: int = 2,
    dropout_probability: float = 0.5,
    z_score_theta: Optional[str] = "independent",
    embedding_net_theta: nn.Module = nn.Identity(),
) -> Callable:
    """
    Return function that will be the neural network when called with parameters.

    We need data to estimate the mean and std used for z-scoring. Once this data is
    provided, we build the neural network with this function.

    Args:
        model: Neural network used to distinguish valid from invalid samples. If it is
            a string, use a pre-configured network of the provided type (either
            mlp or resnet). Alternatively, a function that builds a custom
            neural network can be provided. The function will be called with the
            first batch of parameters (theta,), which can thus be used for shape
            inference and potentially for z-scoring. It needs to return a PyTorch
            `nn.Module` implementing the classifier.
        hidden_features: Number of hidden units of the classifier if `model` is a
            string.
        num_blocks: Number of hidden layers of the classifier if `model` is a
            string.
        dropout_probability: Dropout probability of the classifier if `model` is
            `resnet`.
        z_score_theta: Whether to z-score parameters $\theta$ before passing them into
            the network, can take one of the following:
            - `none`, or None: do not z-score.
            - `independent`: z-score each dimension independently.
            - `structured`: treat dimensions as related, therefore compute mean and std
            over the entire batch, instead of per-dimension. Should be used when each
            sample is, for example, a time series or an image.
        embedding_net_theta: Neural network used to encode the parameters before they
            are passed to the classifier.

    Returns:
        Function that, when called with a batch of parameters, builds the classifier.
    """
    if model == "resnet":

        def build_nn(theta) -> nn.Module:
            classifier = nets.ResidualNet(
                in_features=theta.shape[1],
                out_features=2,
                hidden_features=hidden_features,
                context_features=None,
                num_blocks=num_blocks,
                activation=relu,
                dropout_probability=dropout_probability,
                use_batch_norm=True,
            )
            input_layer = build_input_layer(theta, z_score_theta, embedding_net_theta)
            classifier = nn.Sequential(input_layer, classifier)
            return classifier

    elif model == "mlp":

        def build_nn(theta) -> nn.Module:
            classifier = nn.Sequential(
                nn.Linear(theta.shape[1], hidden_features),
                nn.BatchNorm1d(hidden_features),
                nn.ReLU(),
                nn.Linear(hidden_features, hidden_features),
                nn.BatchNorm1d(hidden_features),
                nn.ReLU(),
                nn.Linear(hidden_features, 2),
            )
            input_layer = build_input_layer(theta, z_score_theta, embedding_net_theta)
            classifier = nn.Sequential(input_layer, classifier)
            return classifier

    else:
        raise NameError(
            f"The `model` must be either of [resnet|mlp]. You passed {model}."
        )

    return build_nn


class RestrictionEstimator:
    def __init__(
        self,
        prior: Distribution,
        model: Union[str, Callable] = "resnet",
        decision_criterion: Union[str, Callable] = "nan",
        hidden_features: int = 100,
        num_blocks: int = 2,
        dropout_probability: float = 0.5,
        z_score: Optional[str] = "independent",
        embedding_net: nn.Module = nn.Identity(),
    ) -> None:
        r"""
        Estimator that trains a classifier to restrict the prior.

        The classifier learns to distinguish `valid` simulation outputs from `invalid`
        simulation outputs.

        Args:
            prior: Prior distribution.
            model: Neural network used to distinguish valid from invalid samples. If it
                is a string, use a pre-configured network of the provided type (either
                mlp or resnet). Alternatively, a function that builds a custom
                neural network can be provided. The function will be called with the
                first batch of parameters (theta,), which can thus be used for shape
                inference and potentially for z-scoring. It needs to return a PyTorch
                `nn.Module` implementing the classifier.
            decision_criterion: Callable that takes in the simulation output $x$ and
                outputs whether $x$ is counted as `valid` simulation (output 1) or as a
                `invalid` simulation (output 0). By default, the function checks
                whether a simulation output $x$ contains at least one `nan` or `inf`.
            hidden_features: Number of hidden units of the classifier if `model` is a
                string.
            num_blocks: Number of hidden layers of the classifier if `model` is a
                string.
            dropout_probability: Dropout probability of the classifier if `model` is
                `resnet`.
            z_score: Whether to z-score the parameters $\theta$ used to train the
                classifier.
            embedding_net: Neural network used to encode the parameters before they are
                passed to the classifier.
        """
        self._prior = prior
        self._classifier = None

        self._device = "cpu"  # TODO hot fix to prevent the tests from crashing

        if isinstance(model, str):
            build_nn = build_classifier(
                model,
                hidden_features,
                num_blocks,
                dropout_probability,
                z_score,
                embedding_net,
            )
        else:
            build_nn = model

        self._build_nn = build_nn
        self._valid_or_invalid_criterion = decision_criterion

        self._theta_roundwise = []
        self._x_roundwise = []
        self._label_roundwise = []
        self._data_round_index = []
        self._validation_log_probs = []

    def append_simulations(self, theta: Tensor, x: Tensor) -> "RestrictionEstimator":
        r"""
        Store parameters and simulation outputs to use them for training later.
        Data ar stored as entries in lists for each type of variable (parameter/data).

        Args:
            theta: Parameter sets.
            x: Simulation outputs.

        Returns:
            `RestrictionEstimator` object (returned so that this function is chainable).
        """

        theta, x = validate_theta_and_x(theta, x, training_device=self._device)

        if self._valid_or_invalid_criterion == "nan":
            label, _, _ = handle_invalid_x(x)
        else:
            assert isinstance(self._valid_or_invalid_criterion, Callable)
            label = self._valid_or_invalid_criterion(x)

        label = label.long()

        if self._data_round_index:
            self._data_round_index.append(self._data_round_index[-1] + 1)
        else:
            self._data_round_index.append(0)

        self._theta_roundwise.append(theta)
        self._x_roundwise.append(x)
        self._label_roundwise.append(label)

        return self

    def get_simulations(self, starting_round: int = 0) -> Tuple[Tensor, Tensor, Tensor]:
        r"""
        Return all $(\theta, x, label)$ pairs that have been passed to this object.

        The label had been inferred from the `valid_or_invalid_criterion`.
        """
        theta = get_simulations_since_round(
            self._theta_roundwise, self._data_round_index, starting_round
        )
        x = get_simulations_since_round(
            self._x_roundwise, self._data_round_index, starting_round
        )
        label = get_simulations_since_round(
            self._label_roundwise, self._data_round_index, starting_round
        )

        return theta, x, label

    def train(
        self,
        training_batch_size: int = 50,
        learning_rate: float = 5e-4,
        validation_fraction: float = 0.1,
        stop_after_epochs: int = 20,
        max_num_epochs: int = 2**31 - 1,
        clip_max_norm: Optional[float] = 5.0,
        loss_importance_weights: Union[bool, float] = False,
        subsample_invalid_sims: Union[float, str] = 1.0,
    ) -> torch.nn.Module:
        r"""
        Train the classifier to distinguish parameters with `valid`|`invalid` outputs.

        Args:
            training_batch_size: Training batch size.
            learning_rate: Learning rate for Adam optimizer.
            validation_fraction: The fraction of data to use for validation.
            stop_after_epochs: The number of epochs to wait for improvement on the
                validation set before terminating training.
            max_num_epochs: Maximum number of epochs to run. If reached, we stop
                training even when the validation loss is still decreasing. If None, we
                train until validation loss increases (see also `stop_after_epochs`).
            clip_max_norm: Value at which to clip the total gradient norm in order to
                prevent exploding gradients. Use None for no clipping.
            loss_importance_weights: If `bool`: whether or not to reweigh the loss such
                that the prior between `valid` and `invalid` simulations is uniform.
                This is one way to deal with imbalanced data (e.g. 99% invalid
                simulations). If you want to reweigh with a custom weight, pass a
                `float`. The value assigned will be the reweighing factor for invalid
                simulations, (1-reweigh_factor) will be the factor for good simulations.
            subsample_invalid_sims: Sampling weight of invalid simulations. This can be
                useful when the fraction of invalid simulations is extremely high and
                one wants to train on a larger fraction of valid simulations. This
                factor has to be in [0, 1]. If it is `auto`, automatically infer
                subsample weights such that the data is balanced.
        """

        theta: Tensor = torch.cat(self._theta_roundwise)
        label: Tensor = torch.cat(self._label_roundwise)

        # Get indices for permutation of the data.
        num_examples = len(theta)
        permuted_indices = torch.randperm(num_examples)
        num_training_examples = int((1 - validation_fraction) * num_examples)
        num_validation_examples = num_examples - num_training_examples
        train_indices, val_indices = (
            permuted_indices[:num_training_examples],
            permuted_indices[num_training_examples:],
        )

        # The ratio of `valid` and `invalid` simulation outputs might not be balanced.
        # E.g. if there are fewer `valid` datapoints, one might want to sample them
        # more often (i.e. show them to the neural network more often). Such a sampler
        # is implemented below. Also see: https://discuss.pytorch.org/t/29907
        subsample_weights: Tensor = torch.ones(num_examples)
        if subsample_invalid_sims == "auto":
            subsample_invalid = float(label.sum()) / float(theta.shape[0] - label.sum())
        else:
            assert isinstance(subsample_invalid_sims, float)
            subsample_invalid = subsample_invalid_sims

        subsample_weights[torch.logical_not(label.bool())] = subsample_invalid

        subsample_weights = deepcopy(subsample_weights)
        subsample_weights[val_indices] = 0.0

        # Dataset is shared for training and validation loaders.
        dataset = data.TensorDataset(theta, label)

        # Create neural_net and validation loaders using a subset sampler.
        train_loader = data.DataLoader(
            dataset,
            batch_size=training_batch_size,
            drop_last=True,
            sampler=WeightedRandomSampler(
                subsample_weights.tolist(),
                int(subsample_weights.sum()),
                replacement=False,
            ),
        )
        val_loader = data.DataLoader(
            dataset,
            batch_size=min(
                max(200, training_batch_size), num_examples - num_training_examples
            ),
            shuffle=False,
            drop_last=True,
            sampler=SubsetRandomSampler(val_indices.tolist()),
        )

        if self._classifier is None:
            self._classifier = self._build_nn(theta[train_indices])

        # If we are in the first round, save the validation data in order to be able to
        # tune the classifier threshold.
        if max(self._data_round_index) == 0:
            self._first_round_validation_theta = theta[val_indices]
            self._first_round_validation_label = label[val_indices]

        optimizer = optim.Adam(
            list(self._classifier.parameters()),
            lr=learning_rate,
        )

        # Compute the fraction of good simulations in dataset.
        if loss_importance_weights:
            if isinstance(loss_importance_weights, bool):
                good_sim_fraction = torch.sum(label, dtype=torch.float) / label.shape[0]
                importance_weights = good_sim_fraction
            else:
                importance_weights = loss_importance_weights
        else:
            importance_weights = 0.5

        # Factor of two such that the average learning rate remains the same.
        # Needed because the average of reweigh_factor and 1-reweigh_factor will be 0.5
        # only.
        importance_weights = 2 * torch.tensor(
            [importance_weights, 1 - importance_weights]
        )

        criterion = nn.CrossEntropyLoss(importance_weights, reduction="none")

        epoch, self._val_log_prob = 0, float("-Inf")
        while epoch <= max_num_epochs and not self._converged(epoch, stop_after_epochs):
            self._classifier.train()
            for parameters, observations in train_loader:
                optimizer.zero_grad()
                outputs = self._classifier(parameters)
                loss = criterion(outputs, observations).mean()
                loss.backward()
                if clip_max_norm is not None:
                    clip_grad_norm_(
                        self._classifier.parameters(),
                        max_norm=clip_max_norm,
                    )
                optimizer.step()

            epoch += 1

            # calculate validation performance
            self._classifier.eval()

            val_loss = 0.0
            with torch.no_grad():
                for parameters, observations in val_loader:
                    outputs = self._classifier(parameters)
                    loss = criterion(outputs, observations)
                    loss[~observations.bool()] *= subsample_invalid_sims
                    val_loss += loss.sum().item()
            self._val_log_prob = -val_loss / num_validation_examples
            self._validation_log_probs.append(self._val_log_prob)

            print("Training neural network. Epochs trained: ", epoch, end="\r")

        return deepcopy(self._classifier)

    def restrict_prior(
        self,
        classifier: Optional[nn.Module] = None,
        allowed_false_negatives: float = 0.0,
    ) -> "RestrictedPrior":
        r"""
        Return the restricted prior.

        The restricted prior (Deistler et al. 2020, in preparation) is the part of the
        prior that can produce `valid` simulations. More formally, the restricted prior
        $p_r(\theta)$ is:

        $p_r(\theta) = c \cdot p(\theta) if \theta \in support(p(\theta|x=`valid`))$
        $p_r(\theta) = 0 otherwise$.

        We sample from the restricted prior by sampling from the prior and then
        rejecting if the classifier predicts that the simulation output can not be
        `valid`.

        Args:
            classifier: Classifier that is used to predict whether parameter sets are
                `valid` or `invalid`.
            allowed_false_negatives: Fraction of false-negative predictions the
                classifier is allowed to make. The threshold of the classifier will be
                tuned such that this criterion is fulfilled. A high value will lead to
                the classifier rejecting more parameter sets, which will give many
                `valid` parameter sets. However, a high value also means that some
                potentially `valid` parameter sets will be missed. Inference is only
                **exact** for `allowed_false_negatives=0.0`. The value specified here
                corresponds approximately to the fraction of parameter sets that will
                be systematically missed by the inference procedure.

        Returns:
            Restricted prior with `.sample()` and `.predict()` methods.

        """
        if classifier is None:
            assert self._classifier is not None, "Classifier must be trained first."
            classifier_ = self._classifier
        else:
            classifier_ = classifier

        classifier_.zero_grad(set_to_none=True)

        return RestrictedPrior(
            self._prior,
            classifier_,
            self._first_round_validation_theta,
            self._first_round_validation_label,
            allowed_false_negatives,
        )

    def _converged(self, epoch: int, stop_after_epochs: int) -> bool:
        r"""
        Return whether the training converged yet and save best model state so far.

        Checks for improvement in validation performance over previous epochs.

        Args:
            epoch: Current epoch in training.
            stop_after_epochs: How many fruitless epochs to let pass before stopping.

        Returns:
            Whether the training has stopped improving, i.e. has converged.
        """
        converged = False

        assert self._classifier is not None, "Classifier must be trained first."
        posterior_nn = self._classifier

        # (Re)-start the epoch count with the first epoch or any improvement.
        if epoch == 0 or self._val_log_prob > self._best_val_log_prob:
            self._best_val_log_prob = self._val_log_prob
            self._epochs_since_last_improvement = 0
            self._best_model_state_dict = deepcopy(posterior_nn.state_dict())
        else:
            self._epochs_since_last_improvement += 1

        # If no validation improvement over many epochs, stop training.
        if self._epochs_since_last_improvement > stop_after_epochs - 1:
            posterior_nn.load_state_dict(self._best_model_state_dict)
            converged = True

        return converged


class RestrictedPrior:
    def __init__(
        self,
        prior: Distribution,
        classifier: nn.Module,
        validation_theta: Tensor,
        validation_label: Tensor,
        allowed_false_negatives: float = 0.0,
    ) -> None:
        r"""
        Initialize the simulation informed prior.

        Args:
            prior: Prior distribution, will be used as proposal distribution whose
                samples will be evaluated by the classifier.
            classifier: Classifier that is evaluated to check if a parameter set
                $\theta$ is valid or not.
            validation_theta: The parameters in the latest validation set that the
                classifier was trained on. Used to calibrate the classifier threshold.
            validation_label: The labels in the latest validation set that the
                classifier was trained on. Used to calibrate the classifier threshold.
        """

        self._prior = prior
        self._classifier = classifier
        self._validation_theta = validation_theta
        self._validation_label = validation_label
        self._classifier_thr = None
        self._reweigh_factor = None
        self.acceptance_rate = None

        self.tune_rejection_threshold(allowed_false_negatives)

    def classifier_probs(self, theta: Tensor) -> Tensor:
        r"""
        Return probability that the parameter set produces a `valid` simulation-output.

        Args:
            theta: Parameters whose label to predict.

        Returns:
            Probability that the parameter set is `valid'.
        """
        pred = F.softmax(self._classifier.forward(theta), dim=1)[:, 1]
        return pred

    def sample(
        self,
        sample_shape: Shape = torch.Size(),
        show_progress_bars: bool = False,
        max_sampling_batch_size: int = 10_000,
        save_acceptance_rate: bool = False,
    ) -> Tensor:
        """
        Return samples from the `RestrictedPrior`.

        Samples are obtained by sampling from the prior, evaluating them under the
        trained classifier (`RestrictionEstimator`) and using only those that were
        accepted.

        Args:
            sample_shape: Shape of the returned samples.
            show_progress_bars: Whether or not to show a progressbar during sampling.
            max_sampling_batch_size: Batch size for drawing samples from the posterior.
                Takes effect only in the second iteration of the loop below, i.e., in
                case of leakage or `num_samples>max_sampling_batch_size`. Larger batch
                size speeds up sampling.
            save_acceptance_rate: If `True`, the acceptance rate is saved and such that
                it can potentially be used later in `log_prob()`.

        Returns:
            Samples from the `RestrictedPrior`.
        """

        num_samples = torch.Size(sample_shape).numel()
        num_sampled_total, num_remaining = 0, num_samples
        accepted, acceptance_rate = [], float("Nan")

        # Progress bar can be skipped.
        pbar = tqdm(
            disable=not show_progress_bars,
            total=num_samples,
            desc=f"Drawing {num_samples} posterior samples",
        )

        # To cover cases with few samples without leakage:
        sampling_batch_size = min(num_samples, max_sampling_batch_size)
        while num_remaining > 0:
            # Sample and reject.
            candidates = self._prior.sample(torch.Size((sampling_batch_size,))).reshape(
                sampling_batch_size, -1
            )
            are_accepted_by_classifier = self.predict(candidates)
            samples = candidates[are_accepted_by_classifier.bool()]
            accepted.append(samples)

            # Update.
            num_sampled_total += sampling_batch_size
            num_remaining -= samples.shape[0]
            pbar.update(samples.shape[0])

            # To avoid endless sampling when leakage is high, we raise a warning if the
            # acceptance rate is too low after the first 1_000 samples.
            acceptance_rate = (num_samples - num_remaining) / num_sampled_total

            # For remaining iterations (leakage or many samples) continue sampling with
            # fixed batch size.
            sampling_batch_size = max_sampling_batch_size

        if save_acceptance_rate:
            self.acceptance_rate = torch.as_tensor(acceptance_rate)

        pbar.close()
        print(
            f"The classifier rejected {(1.0 - acceptance_rate) * 100:.1f}% of all "
            f"samples. You will get a speed-up of "
            f"{(1.0 / acceptance_rate - 1.0) * 100:.1f}%.",
        )

        # When in case of leakage a batch size was used there could be too many samples.
        samples = torch.cat(accepted)[:num_samples]
        assert (
            samples.shape[0] == num_samples
        ), "Number of accepted samples must match required samples."

        return samples

    def log_prob(
        self,
        theta: Tensor,
        norm_restricted_prior: bool = True,
        track_gradients: bool = False,
        prior_acceptance_params: Optional[dict] = None,
    ) -> Tensor:
        r"""Returns the log-probability of the restricted prior.

        Args:
            theta: Parameters $\theta$.
            norm_restricted_prior: Whether to enforce a normalized restricted prior
                density. The normalizing factor is calculated via rejection sampling,
                so if you need speedier but unnormalized log probability estimates set
                here `norm_restricted_prior=False`. The returned log probability is set
                to -∞ outside of the restriceted prior support regardless of this
                setting.
            track_gradients: Whether the returned tensor supports tracking gradients.
                This can be helpful for e.g. sensitivity analysis, but increases memory
                consumption.
            prior_acceptance_params: A `dict` of keyword arguments to override the
                default values of `prior_acceptance()`. Possible options are:
                `num_rejection_samples`, `force_update`, `show_progress_bars`, and
                `rejection_sampling_batch_size`. These parameters only have an effect
                if `norm_restricted_prior=True`.

        Returns:
            `(len(θ),)`-shaped log probability for θ in the support of the restricted
            prior, -∞ (corresponding to 0 probability) outside.
        """
        theta = ensure_theta_batched(torch.as_tensor(theta))

        with torch.set_grad_enabled(track_gradients):

            # Evaluate on device, move back to cpu for comparison with prior.
            prior_log_prob = self._prior.log_prob(theta)
            accepted_by_classifer = self.predict(theta)

            masked_log_prob = torch.where(
                accepted_by_classifer.bool(),
                prior_log_prob,
                torch.tensor(float("-inf"), dtype=torch.float32),
            )

            if prior_acceptance_params is None:
                prior_acceptance_params = dict()  # use defaults
            log_factor = (
                torch.log(self.prior_acceptance(**prior_acceptance_params))
                if norm_restricted_prior
                else 0
            )

            return masked_log_prob - log_factor

    @torch.no_grad()
    def prior_acceptance(
        self,
        num_rejection_samples: int = 10_000,
        force_update: bool = False,
        show_progress_bars: bool = False,
        rejection_sampling_batch_size: int = 10_000,
    ) -> Tensor:
        r"""Return the fraction of prior samples accepted by the classifier.

        The factor is estimated from the acceptance probability during rejection
        sampling from the prior.

        Arguments:
            num_rejection_samples: Number of samples used to estimate correction factor.
            show_progress_bars: Whether to show a progress bar during sampling.
            rejection_sampling_batch_size: Batch size for rejection sampling.

        Returns:
            Estimated acceptance rate.
        """
        if self.acceptance_rate is None or force_update:
            _ = self.sample(
                (num_rejection_samples,),
                show_progress_bars=show_progress_bars,
                max_sampling_batch_size=rejection_sampling_batch_size,
                save_acceptance_rate=True,
            )
        return self.acceptance_rate  # type:ignore

    def predict(self, theta: Tensor) -> Tensor:
        r"""
        Run classifier to predict whether the parameter set is `invalid` or `valid`.

        Args:
            theta: Parameters whose label to predict.

        Returns:
            Integers that indicate whether the parameter set is predicted to be
            `invalid` (=0) or `valid` (=1).
        """

        pred = self.classifier_probs(theta)
        if self._reweigh_factor is None:
            threshold = self._classifier_thr
            predictions = self.classifier_probs(theta) > threshold
        else:
            probs_invalid = pred * self._reweigh_factor
            probs_valid = (1 - pred) * (1 - self._reweigh_factor)
            predictions = probs_valid > probs_invalid
        return predictions.int()

    def tune_rejection_threshold(
        self,
        allowed_false_negatives: Optional[float] = None,
        reweigh_factor: Optional[float] = None,
        print_fp_rate=False,
        safety_margin: Optional[Union[str, float]] = "frequentist",
    ) -> None:
        r"""
        Set the decision threshold of the classifier.

        Compute the highest decision threshold at which the number of false negatives
        is smaller than `allowed_false_negatives`. This threshold will then be set as
        default when calling `.forward()` and for `.build_sim_informed_prior()`.

        Args:
            allowed_false_negatives: Allowed fraction of false negatives on a held-out
                test set.
            reweigh_factor: Post-hoc correction factor. Should be in [0, 1]. A large
                reweigh factor will increase the probability of predicting a `invalid`
                simulation.
            print_fp_rate: Whether or not to compute and print the false-positive rate
                at the obtained threshold.
            safety_margin: When `allowed_false_negatives=0.0`, we might want to apply
                an additional `safety_margin` to the threshold. If `None`, there will
                be no margin and the threshold will be the minimum prediction among all
                `valid' parameter sets. If it is a `float`, this float will be
                subtracted from the minimum prediction. Lastly, if it is a `str`, we
                reduce the classifier threshold using a statistical estimator (see
                `German Tank Problem`). Only supported estimator is `frequentist`.
                Small note: the procedure is fully correct only for a uniform
                distribution of classifier predictions.
        """

        assert (
            allowed_false_negatives is None or reweigh_factor is None
        ), """Both the `allowed_false_negatives` and the `reweigh_factor` are set. You
            can only set one of them."""
        self._reweigh_factor = reweigh_factor

        valid_val_theta = self._validation_theta[self._validation_label.bool()]
        num_valid = valid_val_theta.shape[0]
        clf_probs = self.classifier_probs(valid_val_theta)

        if allowed_false_negatives == 0.0:
            if safety_margin is None:
                self._classifier_thr = torch.min(clf_probs)
            elif isinstance(safety_margin, float):
                self._classifier_thr = torch.min(clf_probs) - safety_margin
            elif safety_margin == "frequentist":
                # We seek the minimum classifier output, not the maximum, as it usually
                # is in the `German Tank Problem`. Hence, we transform the outputs with
                # (1-output), apply the estimator, and then transform back.
                tf_min_val = torch.max(1.0 - clf_probs)
                tf_estimate = tf_min_val + tf_min_val / num_valid
                self._classifier_thr = 1.0 - tf_estimate
            else:
                raise NameError(f"`safety_margin` {safety_margin} not supported.")
        else:
            assert allowed_false_negatives is not None, "Set allowed_false_negatives!"
            quantile_index = floor(num_valid * allowed_false_negatives)
            self._classifier_thr, _ = torch.kthvalue(clf_probs, quantile_index + 1)

        self._classifier_thr = self._classifier_thr.detach()
        if print_fp_rate:
            self.print_false_positive_rate()

    def print_false_positive_rate(self) -> float:
        r"""
        Print and return the rate of false positive predictions on the validation set.

        Returns:
            The flase positive rate.
        """
        invalid_val_theta = self._validation_theta[~self._validation_label.bool()]
        predictions = self.predict(invalid_val_theta)
        num_false_positives = int(predictions.sum())
        fraction_false_positives = num_false_positives / invalid_val_theta.shape[0]
        print(
            f"Fraction of false positives: "
            f"{num_false_positives} / {invalid_val_theta.shape[0]} = "
            f"{fraction_false_positives:.3f}"
        )
        return fraction_false_positives



# This file is part of sbi, a toolkit for simulation-based inference. sbi is licensed
# under the Affero General Public License v3, see <https://www.gnu.org/licenses/>.

import warnings
from typing import Optional, Tuple

import torch
from pyknos.nflows import flows
from torch import Tensor, nn, unique
from torch.distributions import Categorical
from torch.nn import Sigmoid, Softmax

from sbi.neural_nets.flow import build_nsf
from sbi.utils.sbiutils import match_theta_and_x_batch_shapes, standardizing_net
from sbi.utils.torchutils import atleast_2d
from sbi.utils.user_input_checks import check_data_device


def build_mnle(
    batch_x: Tensor,
    batch_y: Tensor,
    z_score_x: Optional[str] = "independent",
    z_score_y: Optional[str] = "independent",
    num_transforms: int = 2,
    num_bins: int = 5,
    hidden_features: int = 50,
    hidden_layers: int = 2,
    tail_bound: float = 10.0,
    log_transform_x: bool = True,
    **kwargs,
):
    """Returns a density estimator for mixed data types.

    Uses a categorical net to model the discrete part and a neural spline flow (NSF) to
    model the continuous part of the data.

    Args:
        batch_x: batch of data
        batch_y: batch of parameters
        z_score_x: whether to z-score x.
        z_score_y: whether to z-score y.
        num_transforms: number of transforms in the NSF
        num_bins: bins per spline for NSF.
        hidden_features: number of hidden features used in both nets.
        hidden_layers: number of hidden layers in the categorical net.
        tail_bound: spline tail bound for NSF.
        log_transform_x: whether to apply a log-transform to x to move it to unbounded
            space, e.g., in case x consists of reaction time data (bounded by zero).

    Returns:
        MixedDensityEstimator: nn.Module for performing MNLE.
    """

    check_data_device(batch_x, batch_y)
    if z_score_y == "independent":
        embedding = standardizing_net(batch_y)
    else:
        embedding = None

    warnings.warn(
        """The mixed neural likelihood estimator assumes that x contains
        continuous data in the first n-1 columns (e.g., reaction times) and
        categorical data in the last column (e.g., corresponding choices). If
        this is not the case for the passed `x` do not use this function."""
    )
    # Separate continuous and discrete data.
    cont_x, disc_x = _separate_x(batch_x)

    # Infer input and output dims.
    dim_parameters = batch_y[0].numel()
    num_categories = unique(disc_x).numel()

    # Set up a categorical RV neural net for modelling the discrete data.
    disc_nle = CategoricalNet(
        num_input=dim_parameters,
        num_categories=num_categories,
        num_hidden=hidden_features,
        num_layers=hidden_layers,
        embedding=embedding,
    )

    # Set up a NSF for modelling the continuous data, conditioned on the discrete data.
    cont_nle = build_nsf(
        batch_x=torch.log(cont_x)
        if log_transform_x
        else cont_x,  # log transform manually.
        batch_y=torch.cat((batch_y, disc_x), dim=1),  # condition on discrete data too.
        z_score_y=z_score_y,
        z_score_x=z_score_x,
        num_bins=num_bins,
        num_transforms=num_transforms,
        tail_bound=tail_bound,
        hidden_features=hidden_features,
    )

    return MixedDensityEstimator(
        discrete_net=disc_nle,
        continuous_net=cont_nle,
        log_transform_x=log_transform_x,
    )


class CategoricalNet(nn.Module):
    """Class to perform conditional density (mass) estimation for a categorical RV.

    Takes as input parameters theta and learns the parameters p of a Categorical.

    Defines log prob and sample functions.
    """

    def __init__(
        self,
        num_input: int = 4,
        num_categories: int = 2,
        num_hidden: int = 20,
        num_layers: int = 2,
        embedding: Optional[nn.Module] = None,
    ):
        """Initialize the neural net.

        Args:
            num_input: number of input units, i.e., dimensionality of parameters.
            num_categories: number of output units, i.e., number of categories.
            num_hidden: number of hidden units per layer.
            num_layers: number of hidden layers.
            embedding: emebedding net for parameters, e.g., a z-scoring transform.
        """
        super(CategoricalNet, self).__init__()

        self.num_hidden = num_hidden
        self.num_input = num_input
        self.activation = Sigmoid()
        self.softmax = Softmax(dim=1)
        self.num_categories = num_categories

        # Maybe add z-score embedding for parameters.
        if embedding is not None:
            self.input_layer = nn.Sequential(
                embedding, nn.Linear(num_input, num_hidden)
            )
        else:
            self.input_layer = nn.Linear(num_input, num_hidden)

        # Repeat hidden units hidden layers times.
        self.hidden_layers = nn.ModuleList()
        for _ in range(num_layers):
            self.hidden_layers.append(nn.Linear(num_hidden, num_hidden))

        self.output_layer = nn.Linear(num_hidden, num_categories)

    def forward(self, theta: Tensor) -> Tensor:
        """Return categorical probability predicted from a batch of parameters.

        Args:
            theta: batch of input parameters for the net.

        Returns:
            Tensor: batch of predicted categorical probabilities.
        """
        assert theta.dim() == 2, "input needs to have a batch dimension."
        assert (
            theta.shape[1] == self.num_input
        ), f"input dimensions must match num_input {self.num_input}"

        # forward path
        theta = self.activation(self.input_layer(theta))

        # iterate n hidden layers, input x and calculate tanh activation
        for layer in self.hidden_layers:
            theta = self.activation(layer(theta))

        return self.softmax(self.output_layer(theta))

    def log_prob(self, x: Tensor, theta: Tensor) -> Tensor:
        """Return categorical log probability of categories x, given parameters theta.

        Args:
            theta: parameters.
            x: categories to evaluate.

        Returns:
            Tensor: log probs with shape (x.shape[0],)
        """
        # Predict categorical ps and evaluate.
        ps = self.forward(theta)
        return Categorical(probs=ps).log_prob(x.squeeze())

    def sample(self, num_samples: int, theta: Tensor) -> Tensor:
        """Returns samples from categorical random variable with probs predicted from
        the neural net.

        Args:
            theta: batch of parameters for prediction.
            num_samples: number of samples to obtain.

        Returns:
            Tensor: Samples with shape (num_samples, 1)
        """

        # Predict Categorical ps and sample.
        ps = self.forward(theta)
        return (
            Categorical(probs=ps)
            .sample(torch.Size((num_samples,)))
            .reshape(num_samples, -1)
        )


class MixedDensityEstimator(nn.Module):
    """Class performing Mixed Neural Likelihood Estimation.

    MNLE combines a Categorical net and a neural spline flow to model data with
    mixed types, e.g., as they occur in decision-making models.
    """

    def __init__(
        self,
        discrete_net: CategoricalNet,
        continuous_net: flows.Flow,
        log_transform_x: bool = False,
    ):
        """Initialize class for combining density estimators for MNLE.

        Args:
            discrete_net: neural net to model discrete part of the data.
            continuous_net: neural net to model the continuous data.
            log_transform_x: whether to transform the continous part of the data into
                logarithmic domain before training. This is helpful for bounded data, e.
                g.,for reaction times.
        """
        super(MixedDensityEstimator, self).__init__()

        self.discrete_net = discrete_net
        self.continuous_net = continuous_net
        self.log_transform_x = log_transform_x

    def forward(self, x: Tensor):
        raise NotImplementedError(
            """The forward method is not implemented for MNLE, use '.sample(...)' to
            generate samples though a forward pass."""
        )

    def sample(
        self, theta: Tensor, num_samples: int = 1, track_gradients: bool = False
    ) -> Tensor:
        """Return sample from mixed data distribution.

        Args:
            theta: parameters for which to generate samples.
            num_samples: number of samples to generate.

        Returns:
            Tensor: samples with shape (num_samples, num_data_dimensions)
        """
        assert theta.shape[0] == 1, "Samples can be generated for a single theta only."

        with torch.set_grad_enabled(track_gradients):

            # Sample discrete data given parameters.
            discrete_x = self.discrete_net.sample(
                theta=theta,
                num_samples=num_samples,
            ).reshape(num_samples, 1)

            # Sample continuous data condition on parameters and discrete data.
            # Pass num_samples=1 because the choices in the context contains
            # num_samples elements already.
            continuous_x = self.continuous_net.sample(
                # repeat the single theta to match number of sampled choices.
                context=torch.cat((theta.repeat(num_samples, 1), discrete_x), dim=1),
                num_samples=1,
            ).reshape(num_samples, 1)

            # In case x was log-transformed, move them to linear space.
            if self.log_transform_x:
                continuous_x = continuous_x.exp()

        return torch.cat((continuous_x, discrete_x), dim=1)

    def log_prob(self, x: Tensor, context: Tensor) -> Tensor:
        """Return log-probability of samples under the learned MNLE.

        For a fixed data point x this returns the value of the likelihood function
        evaluated at theta, L(theta, x=x).

        Alternatively, it can be interpreted as the log-prob of the density
        p(x | theta).

        It evaluates the separate density estimator for the discrete and continous part
        of the data and then combines them into one evaluation.

        Args:
            x: data (containing continuous and discrete data).
            context: parameters for which to evaluate the likelihod function, or for
                which to condition p(x | theta).

        Returns:
            Tensor: log_prob of p(x | theta).
        """
        assert (
            x.shape[0] == context.shape[0]
        ), "x and context must have same batch size."

        cont_x, disc_x = _separate_x(x)
        num_parameters = context.shape[0]

        disc_log_prob = self.discrete_net.log_prob(x=disc_x, theta=context).reshape(
            num_parameters
        )

        cont_log_prob = self.continuous_net.log_prob(
            # Transform to log-space if needed.
            torch.log(cont_x) if self.log_transform_x else cont_x,
            # Pass parameters and discrete x as context.
            context=torch.cat((context, disc_x), dim=1),
        )

        # Combine into joint lp.
        log_probs_combined = (disc_log_prob + cont_log_prob).reshape(num_parameters)

        # Maybe add log abs det jacobian of RTs: log(1/x) = - log(x)
        if self.log_transform_x:
            log_probs_combined -= torch.log(cont_x).squeeze()

        return log_probs_combined

    def log_prob_iid(self, x: Tensor, theta: Tensor) -> Tensor:
        """Return log prob given a batch of iid x and a different batch of theta.

        This is different from `.log_prob()` to enable speed ups in evaluation during
        inference. The speed up is achieved by exploiting the fact that there are only
        finite number of possible categories in the discrete part of the dat: one can
        just calculate the log probs for each possible category (given the current batch
        of theta) and then copy those log probs into the entire batch of iid categories.
        For example, for the drift-diffusion model, there are only two choices, but
        often 100s or 1000 trials. With this method a evaluation over trials then passes
        a batch of `2 (one per choice) * num_thetas` into the NN, whereas the normal
        `.log_prob()` would pass `1000 * num_thetas`.

        Args:
            x: batch of iid data, data observed given the same underlying parameters or
                experimental conditions.
            theta: batch of parameters to be evaluated, i.e., each batch entry will be
                evaluated for the entire batch of iid x.

        Returns:
            Tensor: log probs with shape (num_trials, num_parameters), i.e., the log
                prob for each theta for each trial.
        """

        theta = atleast_2d(theta)
        x = atleast_2d(x)
        batch_size = theta.shape[0]
        num_trials = x.shape[0]
        theta_repeated, x_repeated = match_theta_and_x_batch_shapes(theta, x)
        net_device = next(self.discrete_net.parameters()).device
        assert (
            net_device == x.device and x.device == theta.device
        ), f"device mismatch: net, x, theta: {net_device}, {x.device}, {theta.device}."

        x_cont_repeated, x_disc_repeated = _separate_x(x_repeated)
        x_cont, x_disc = _separate_x(x)

        log_prob_per_cat = torch.zeros(self.discrete_net.num_categories, batch_size)
        # repeat categories for parameters
        repeated_categories = torch.repeat_interleave(
            torch.arange(self.discrete_net.num_categories - 1), batch_size, dim=0
        )
        # repeat parameters for categories
        repeated_theta = theta.repeat(self.discrete_net.num_categories - 1, 1)
        log_prob_per_cat[:-1, :] = self.discrete_net.log_prob(
            repeated_categories,
            repeated_theta,
        ).reshape(-1, batch_size)
        # infer the last category logprob from sum to one.
        log_prob_per_cat[-1, :] = torch.log(1 - log_prob_per_cat[:-1, :].exp().sum(0))

        # fill in lps for each occurred category
        log_probs_discrete = log_prob_per_cat[
            x_disc.type_as(torch.zeros(1, dtype=torch.long)).squeeze()
        ].reshape(-1)

        # Get repeat discrete data and theta to match in batch shape for flow eval.
        log_probs_cont = self.continuous_net.log_prob(
            torch.log(x_cont_repeated) if self.log_transform_x else x_cont_repeated,
            context=torch.cat((theta_repeated, x_disc_repeated), dim=1),
        )

        # Combine into joint lp with first dim over trials.
        log_probs_combined = (log_probs_discrete + log_probs_cont).reshape(
            num_trials, batch_size
        )

        # Maybe add log abs det jacobian of RTs: log(1/rt) = - log(rt)
        if self.log_transform_x:
            log_probs_combined -= torch.log(x_cont)

        # Return batch over trials as required by SBI potentials.
        return log_probs_combined


def _separate_x(x: Tensor, num_discrete_columns: int = 1) -> Tuple[Tensor, Tensor]:
    """Returns the continuous and discrete part of the given x.

    Assumes the discrete data to live in the last columns of x.
    """

    assert x.ndim == 2, f"x must have two dimensions but has {x.ndim}."

    return x[:, :-num_discrete_columns], x[:, -num_discrete_columns:]



# This file is part of sbi, a toolkit for simulation-based inference. sbi is licensed
# under the Affero General Public License v3, see <https://www.gnu.org/licenses/>.
from typing import Optional, Union

import torch
from pyknos.nflows import flows
from torch import Tensor, log
from torch.distributions import Distribution

from sbi.inference.posteriors.base_posterior import NeuralPosterior
from sbi.inference.potentials.posterior_based_potential import (
    posterior_estimator_based_potential,
)
from sbi.samplers.rejection.rejection import rejection_sample_posterior_within_prior
from sbi.types import Shape
from sbi.utils import check_prior, match_theta_and_x_batch_shapes, within_support
from sbi.utils.torchutils import ensure_theta_batched


class DirectPosterior(NeuralPosterior):
    r"""Posterior $p(\theta|x_o)$ with `log_prob()` and `sample()` methods, only
    applicable to SNPE.<br/><br/>
    SNPE trains a neural network to directly approximate the posterior distribution.
    However, for bounded priors, the neural network can have leakage: it puts non-zero
    mass in regions where the prior is zero. The `DirectPosterior` class wraps the
    trained network to deal with these cases.<br/><br/>
    Specifically, this class offers the following functionality:<br/>
    - correct the calculation of the log probability such that it compensates for the
      leakage.<br/>
    - reject samples that lie outside of the prior bounds.<br/><br/>
    This class can not be used in combination with SNLE or SNRE.
    """

    def __init__(
        self,
        posterior_estimator: flows.Flow,
        prior: Distribution,
        max_sampling_batch_size: int = 10_000,
        device: Optional[str] = None,
        x_shape: Optional[torch.Size] = None,
    ):
        """
        Args:
            prior: Prior distribution with `.log_prob()` and `.sample()`.
            posterior_estimator: The trained neural posterior.
            max_sampling_batch_size: Batchsize of samples being drawn from
                the proposal at every iteration.
            device: Training device, e.g., "cpu", "cuda" or "cuda:0". If None,
                `potential_fn.device` is used.
            x_shape: Shape of a single simulator output. If passed, it is used to check
                the shape of the observed data and give a descriptive error.
        """
        # Because `DirectPosterior` does not take the `potential_fn` as input, it
        # builds it itself. The `potential_fn` and `theta_transform` are used only for
        # obtaining the MAP.
        check_prior(prior)
        potential_fn, theta_transform = posterior_estimator_based_potential(
            posterior_estimator, prior, None
        )

        super().__init__(
            potential_fn=potential_fn,
            theta_transform=theta_transform,
            device=device,
            x_shape=x_shape,
        )

        self.prior = prior
        self.posterior_estimator = posterior_estimator

        self.max_sampling_batch_size = max_sampling_batch_size
        self._leakage_density_correction_factor = None

        self._purpose = """It samples the posterior network and rejects samples that
            lie outside of the prior bounds."""

    def sample(
        self,
        sample_shape: Shape = torch.Size(),
        x: Optional[Tensor] = None,
        max_sampling_batch_size: int = 10_000,
        sample_with: Optional[str] = None,
        show_progress_bars: bool = True,
    ):
        r"""Return samples from posterior distribution $p(\theta|x)$.

        Args:
            sample_shape: Desired shape of samples that are drawn from posterior. If
                sample_shape is multidimensional we simply draw `sample_shape.numel()`
                samples and then reshape into the desired shape.
            sample_with: This argument only exists to keep backward-compatibility with
                `sbi` v0.17.2 or older. If it is set, we instantly raise an error.
            show_progress_bars: Whether to show sampling progress monitor.
        """

        num_samples = torch.Size(sample_shape).numel()
        x = self._x_else_default_x(x)
        max_sampling_batch_size = (
            self.max_sampling_batch_size
            if max_sampling_batch_size is None
            else max_sampling_batch_size
        )

        if sample_with is not None:
            raise ValueError(
                f"You set `sample_with={sample_with}`. As of sbi v0.18.0, setting "
                f"`sample_with` is no longer supported. You have to rerun "
                f"`.build_posterior(sample_with={sample_with}).`"
            )

        samples = rejection_sample_posterior_within_prior(
            posterior_nn=self.posterior_estimator,
            prior=self.prior,
            x=x,
            num_samples=num_samples,
            show_progress_bars=show_progress_bars,
            max_sampling_batch_size=max_sampling_batch_size,
        )[0]
        return samples

    def log_prob(
        self,
        theta: Tensor,
        x: Optional[Tensor] = None,
        norm_posterior: bool = True,
        track_gradients: bool = False,
        leakage_correction_params: Optional[dict] = None,
    ) -> Tensor:
        r"""Returns the log-probability of the posterior $p(\theta|x)$.

        Args:
            theta: Parameters $\theta$.
            norm_posterior: Whether to enforce a normalized posterior density.
                Renormalization of the posterior is useful when some
                probability falls out or leaks out of the prescribed prior support.
                The normalizing factor is calculated via rejection sampling, so if you
                need speedier but unnormalized log posterior estimates set here
                `norm_posterior=False`. The returned log posterior is set to
                -∞ outside of the prior support regardless of this setting.
            track_gradients: Whether the returned tensor supports tracking gradients.
                This can be helpful for e.g. sensitivity analysis, but increases memory
                consumption.
            leakage_correction_params: A `dict` of keyword arguments to override the
                default values of `leakage_correction()`. Possible options are:
                `num_rejection_samples`, `force_update`, `show_progress_bars`, and
                `rejection_sampling_batch_size`.
                These parameters only have an effect if `norm_posterior=True`.

        Returns:
            `(len(θ),)`-shaped log posterior probability $\log p(\theta|x)$ for θ in the
            support of the prior, -∞ (corresponding to 0 probability) outside.
        """
        x = self._x_else_default_x(x)

        # TODO Train exited here, entered after sampling?
        self.posterior_estimator.eval()

        theta = ensure_theta_batched(torch.as_tensor(theta))
        theta_repeated, x_repeated = match_theta_and_x_batch_shapes(theta, x)

        with torch.set_grad_enabled(track_gradients):

            # Evaluate on device, move back to cpu for comparison with prior.
            unnorm_log_prob = self.posterior_estimator.log_prob(
                theta_repeated, context=x_repeated
            )

            # Force probability to be zero outside prior support.
            in_prior_support = within_support(self.prior, theta_repeated)

            masked_log_prob = torch.where(
                in_prior_support,
                unnorm_log_prob,
                torch.tensor(float("-inf"), dtype=torch.float32, device=self._device),
            )

            if leakage_correction_params is None:
                leakage_correction_params = dict()  # use defaults
            log_factor = (
                log(self.leakage_correction(x=x, **leakage_correction_params))
                if norm_posterior
                else 0
            )

            return masked_log_prob - log_factor

    @torch.no_grad()
    def leakage_correction(
        self,
        x: Tensor,
        num_rejection_samples: int = 10_000,
        force_update: bool = False,
        show_progress_bars: bool = False,
        rejection_sampling_batch_size: int = 10_000,
    ) -> Tensor:
        r"""Return leakage correction factor for a leaky posterior density estimate.

        The factor is estimated from the acceptance probability during rejection
        sampling from the posterior.

        This is to avoid re-estimating the acceptance probability from scratch
        whenever `log_prob` is called and `norm_posterior=True`. Here, it
        is estimated only once for `self.default_x` and saved for later. We
        re-evaluate only whenever a new `x` is passed.

        Arguments:
            num_rejection_samples: Number of samples used to estimate correction factor.
            show_progress_bars: Whether to show a progress bar during sampling.
            rejection_sampling_batch_size: Batch size for rejection sampling.

        Returns:
            Saved or newly-estimated correction factor (as a scalar `Tensor`).
        """

        def acceptance_at(x: Tensor) -> Tensor:

            return rejection_sample_posterior_within_prior(
                posterior_nn=self.posterior_estimator,
                prior=self.prior,
                x=x.to(self._device),
                num_samples=num_rejection_samples,
                show_progress_bars=show_progress_bars,
                sample_for_correction_factor=True,
                max_sampling_batch_size=rejection_sampling_batch_size,
            )[1]

        # Check if the provided x matches the default x (short-circuit on identity).
        is_new_x = self.default_x is None or (
            x is not self.default_x and (x != self.default_x).any()
        )

        not_saved_at_default_x = self._leakage_density_correction_factor is None

        if is_new_x:  # Calculate at x; don't save.
            return acceptance_at(x)
        elif not_saved_at_default_x or force_update:  # Calculate at default_x; save.
            assert self.default_x is not None
            self._leakage_density_correction_factor = acceptance_at(self.default_x)

        return self._leakage_density_correction_factor  # type: ignore

    def map(
        self,
        x: Optional[Tensor] = None,
        num_iter: int = 1_000,
        num_to_optimize: int = 100,
        learning_rate: float = 0.01,
        init_method: Union[str, Tensor] = "posterior",
        num_init_samples: int = 1_000,
        save_best_every: int = 10,
        show_progress_bars: bool = False,
        force_update: bool = False,
    ) -> Tensor:
        r"""Returns the maximum-a-posteriori estimate (MAP).

        The method can be interrupted (Ctrl-C) when the user sees that the
        log-probability converges. The best estimate will be saved in `self._map` and
        can be accessed with `self.map()`. The MAP is obtained by running gradient
        ascent from a given number of starting positions (samples from the posterior
        with the highest log-probability). After the optimization is done, we select the
        parameter set that has the highest log-probability after the optimization.

        Warning: The default values used by this function are not well-tested. They
        might require hand-tuning for the problem at hand.

        For developers: if the prior is a `BoxUniform`, we carry out the optimization
        in unbounded space and transform the result back into bounded space.

        Args:
            x: Deprecated - use `.set_default_x()` prior to `.map()`.
            num_iter: Number of optimization steps that the algorithm takes
                to find the MAP.
            learning_rate: Learning rate of the optimizer.
            init_method: How to select the starting parameters for the optimization. If
                it is a string, it can be either [`posterior`, `prior`], which samples
                the respective distribution `num_init_samples` times. If it is a
                tensor, the tensor will be used as init locations.
            num_init_samples: Draw this number of samples from the posterior and
                evaluate the log-probability of all of them.
            num_to_optimize: From the drawn `num_init_samples`, use the
                `num_to_optimize` with highest log-probability as the initial points
                for the optimization.
            save_best_every: The best log-probability is computed, saved in the
                `map`-attribute, and printed every `save_best_every`-th iteration.
                Computing the best log-probability creates a significant overhead
                (thus, the default is `10`.)
            show_progress_bars: Whether or not to show a progressbar for sampling from
                the posterior.
            force_update: Whether to re-calculate the MAP when x is unchanged and
                have a cached value.
            log_prob_kwargs: Will be empty for SNLE and SNRE. Will contain
                {'norm_posterior': True} for SNPE.

        Returns:
            The MAP estimate.
        """
        return super().map(
            x=x,
            num_iter=num_iter,
            num_to_optimize=num_to_optimize,
            learning_rate=learning_rate,
            init_method=init_method,
            num_init_samples=num_init_samples,
            save_best_every=save_best_every,
            show_progress_bars=show_progress_bars,
            force_update=force_update,
        )



from sbi.analysis.conditional_density import (
    ConditionedMDN,
    conditional_corrcoeff,
    conditonal_potential,
    eval_conditional_density,
)
from sbi.analysis.plot import (
    conditional_marginal_plot,
    conditional_pairplot,
    marginal_plot,
    pairplot,
    sbc_rank_plot,
)
from sbi.analysis.sbc import check_sbc, get_nltp, run_sbc
from sbi.analysis.sensitivity_analysis import ActiveSubspace
from sbi.analysis.tensorboard_output import list_all_logs, plot_summary



# This file is part of sbi, a toolkit for simulation-based inference. sbi is licensed
# under the Affero General Public License v3, see <https://www.gnu.org/licenses/>.

from __future__ import annotations

import pytest
import torch
from torch import eye, ones, zeros

from sbi import analysis as analysis
from sbi import utils as utils
from sbi.inference import SNLE, SNPE, SNRE
from sbi.neural_nets.embedding_nets import FCEmbedding
from sbi.simulators.linear_gaussian import linear_gaussian
from sbi.utils import classifier_nn, likelihood_nn, posterior_nn


@pytest.mark.parametrize("method", ["SNPE", "SNLE", "SNRE"])
@pytest.mark.parametrize("num_dim", [1, 2])
@pytest.mark.parametrize("embedding_net", ["mlp"])
def test_embedding_net_api(method, num_dim: int, embedding_net: str):
    """Tests the API when using a preconfigured embedding net."""

    x_o = zeros(1, num_dim)

    # likelihood_mean will be likelihood_shift+theta
    likelihood_shift = -1.0 * ones(num_dim)
    likelihood_cov = 0.3 * eye(num_dim)

    prior = utils.BoxUniform(-2.0 * ones(num_dim), 2.0 * ones(num_dim))

    theta = prior.sample((1000,))
    x = linear_gaussian(theta, likelihood_shift, likelihood_cov)

    if embedding_net == "mlp":
        embedding = FCEmbedding(input_dim=num_dim)
    else:
        raise NameError

    if method == "SNPE":
        density_estimator = posterior_nn("maf", embedding_net=embedding)
        inference = SNPE(
            prior, density_estimator=density_estimator, show_progress_bars=False
        )
    elif method == "SNLE":
        density_estimator = likelihood_nn("maf", embedding_net=embedding)
        inference = SNLE(
            prior, density_estimator=density_estimator, show_progress_bars=False
        )
    elif method == "SNRE":
        classifier = classifier_nn("resnet", embedding_net_x=embedding)
        inference = SNRE(prior, classifier=classifier, show_progress_bars=False)
    else:
        raise NameError

    _ = inference.append_simulations(theta, x).train(max_num_epochs=5)
    posterior = inference.build_posterior().set_default_x(x_o)

    s = posterior.sample((1,))
    _ = posterior.potential(s)
